from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.utils.math_functions import keops_mod


class Mod(VectorizedScalarOp):
    """the Modulo vectorized operation
    Mod(x,n,d) = x - n * Floor((x - d)/n)
    """

    string_id = "Mod"

    ScalarOpFun = keops_mod

    def DiffT(self, v, gradin):
        from keopscore.formulas.maths.Floor import Floor

        # we fall back to an alternative definition of Mod for defining the gradient
        x, n, d = self.children
        Mod_alt = x - n * Floor((x - d) / n)
        return Mod_alt.DiffT(v, gradin)

    # parameters for testing the operation (optional)
    nargs = 3  # number of arguments


"""
# N.B. below is alternative definition as a simple alias.
# It is theoretically less efficient when applied to vectors
# because we compose operations so evalauation implies the creation
# of useless temporary vectors. 
# If we implement a sort of "fusion" of vectorized scalar operations,
# it should be completely equivalent.
        
from keopscore.formulas.maths.Floor import Floor
def Mod(x,n,d):
    return x - n * Floor((x - d)/n)
"""



from keopscore.formulas.Operation import Operation
from keopscore.utils.code_gen_utils import c_for_loop
from keopscore.formulas.complex.Real2Complex import Real2Complex
from keopscore.formulas.complex.ComplexMult import ComplexMult
from keopscore.utils.misc_utils import KeOps_Error

# /////////////////////////////////////////////////////////////////////////
# ////      ComplexRealScal                           ////
# /////////////////////////////////////////////////////////////////////////


class ComplexRealScal_Impl(Operation):

    string_id = "ComplexRealScal"

    def __init__(self, f, g):
        if f.dim != 1:
            KeOps_Error("Dimension of F must be 1")
        if g.dim % 2 != 0:
            KeOps_Error("Dimension of G must be even")
        self.dim = g.dim
        super().__init__(f, g)

    def Op(self, out, table, inF, inG):
        forloop, i = c_for_loop(0, self.dim, 2, pragma_unroll=True)
        body = out[i].assign(inF[0] * inG[i])
        body += out[i + 1].assign(inF[0] * inG[i + 1])
        return forloop(body)

    def DiffT(self, v, gradin):
        f, g = self.children
        AltFormula = ComplexMult(Real2Complex(f), g)
        return AltFormula.DiffT(v, gradin)


def ComplexRealScal(f, g):
    return ComplexRealScal_Impl(f, g) if f.dim == 1 else ComplexRealScal_Impl(g, f)



from keopscore.formulas.maths.Square import Square
from keopscore.formulas.maths.Sum import Sum


class SqNormDiag:
    """
    Anisotropic (but diagonal) norm, if S.dim == A.dim: SqNormDiag(S,A) = sum_i s_i*a_i*a_i
    """

    def __new__(cls, S, A):
        return Sum(S * Square(A))

    enable_test = False



from keopscore.formulas.maths.Inv import Inv
from keopscore.formulas.variables.IntCst import IntCst


class IntInv:
    def __new__(cls, arg):
        return Inv(IntCst(arg))

    enable_test = False



from keopscore.utils.code_gen_utils import (
    infinity,
    c_zero_float,
    VectApply,
    c_if,
)
from keopscore.formulas.reductions.Reduction import Reduction
from keopscore.utils.misc_utils import KeOps_Error


class Min_ArgMin_Reduction_Base(Reduction):
    """min+argmin reduction : base class"""

    def __init__(self, formula, tagIJ):
        super().__init__(formula, tagIJ)

        # We work with a (values,indices) vector
        self.dimred = 2 * formula.dim  # dimension of inner reduction variables

    def InitializeReduction(self, acc):
        # Returns C++ code to be used at initialization phase of the reduction.
        dim = self.formula.dim
        acc_min, acc_argmin = acc.split(dim, dim)
        return acc_min.assign(infinity(acc.dtype)) + acc_argmin.assign(c_zero_float)

    def ReducePairScalar(self, acc_val, acc_ind, xi, ind):
        # Subroutine of ReducePairShort and ReducePair methods.
        if xi.dtype == "half2":
            KeOps_Error("not implemented")
        return c_if(xi < acc_val, acc_val.assign(xi) + acc_ind.assign(ind))

    def ReducePair(self, acc, xi):
        # Returns C++ code that implements the update phase of the reduction.
        dim = self.formula.dim
        acc_val, acc_ind = acc.split(dim, dim)
        xi_val, xi_ind = xi.split(dim, dim)
        return VectApply(self.ReducePairScalar, acc_val, xi_val, xi_ind)

    def ReducePairShort(self, acc, xi, ind):
        if xi.dtype == "half2":
            KeOps_Error("not implemented")
            half2_val = c_variable("half2_ind")
            string = half2_val.declare_assign(
                f"__floats2half2_rn(2*{ind()},2*{ind()}+1)"
            )
        dim = self.formula.dim
        acc_val, acc_ind = acc.split(dim, dim)
        return VectApply(self.ReducePairScalar, acc_val, acc_ind, xi, ind)



from keopscore.formulas.maths.SqNorm2 import SqNorm2


class WeightedSqDist:
    """
    WEIGHTED SQUARED DISTANCE : WeightedSqDist(S,A)
    """

    def __new__(cls, S, A):
        return S * SqNorm2(A)

    enable_test = False



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.formulas.maths.Rsqrt import Rsqrt
from keopscore.utils.math_functions import keops_acos


class Acos(VectorizedScalarOp):
    """the arc-cosine vectorized operation"""

    string_id = "Acos"

    ScalarOpFun = keops_acos

    @staticmethod
    def Derivative(f):
        return -Rsqrt(1 - f**2)

    # parameters for testing the operation (optional)
    test_ranges = [(-1, 1)]  # range of argument



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.utils.math_functions import keops_pow


class Pow(VectorizedScalarOp):
    """the integer power vectorized operation
    Pow(f,m) where m is integer, computes f^m
    """

    def __init__(self, f, m):
        super().__init__(f, params=(m,))

    string_id = "Pow"

    ScalarOpFun = keops_pow

    @staticmethod
    def Derivative(f, m):
        from keopscore.formulas.variables.IntCst import IntCst

        return IntCst(m) * Pow(f, m - 1)

    # parameters for testing the operation (optional)
    nargs = 1  # number of arguments (excluding parameters)
    test_ranges = [(0, 2)]  # ranges of arguments
    test_params = [2]  # values of parameters for testing
    torch_op = "lambda x,m : torch.pow(x, m)"



from keopscore.utils.code_gen_utils import (
    c_array,
    c_zero_float,
    c_if,
    c_variable,
)


class Sum_Scheme:
    def __init__(self, red_formula, dtype, dimred=None):
        self.red_formula = red_formula
        if dimred is None:
            self.dimred = red_formula.dimred
        else:
            self.dimred = dimred

    def declare_temporary_accumulator(self):
        return self.tmp_acc.declare()

    def initialize_temporary_accumulator_first_init(self):
        return ""

    def initialize_temporary_accumulator_block_init(self):
        return ""

    def periodic_accumulate_temporary(self, acc, j):
        return ""

    def final_operation(self, acc):
        return ""


class direct_sum(Sum_Scheme):
    def declare_temporary_accumulator(self):
        return ""

    def initialize_temporary_accumulator(self):
        return ""

    def accumulate_result(self, acc, fout, j, hack=False):
        return self.red_formula.ReducePairShort(acc, fout, j)


class block_sum(Sum_Scheme):
    def __init__(self, red_formula, dtype, dimred=None):
        super().__init__(red_formula, dtype, dimred)
        self.tmp_acc = c_array(dtype, self.dimred, "tmp")

    def initialize_temporary_accumulator(self):
        return (
            "int period_accumulate = ny<10 ? 100 : sqrt(ny);\n"
            + self.red_formula.InitializeReduction(self.tmp_acc)
        )

    def initialize_temporary_accumulator_block_init(self):
        return self.red_formula.InitializeReduction(self.tmp_acc)

    def accumulate_result(self, acc, fout, j, hack=False):
        tmp_acc = acc if hack else self.tmp_acc
        return self.red_formula.ReducePairShort(tmp_acc, fout, j)

    def periodic_accumulate_temporary(self, acc, j):
        condition = c_variable("bool", f"!(({j.id}+1)%period_accumulate)")
        return c_if(
            condition,
            self.red_formula.ReducePair(acc, self.tmp_acc)
            + self.red_formula.InitializeReduction(self.tmp_acc),
        )

    def final_operation(self, acc):
        return self.red_formula.ReducePair(acc, self.tmp_acc)


class kahan_scheme(Sum_Scheme):
    def __init__(self, red_formula, dtype, dimred=None):
        super().__init__(red_formula, dtype, dimred)
        self.tmp_acc = c_array(dtype, red_formula.dim_kahan, "tmp")

    def initialize_temporary_accumulator(self):
        return self.tmp_acc.assign(c_zero_float)

    def initialize_temporary_accumulator_first_init(self):
        return self.initialize_temporary_accumulator()

    def accumulate_result(self, acc, fout, j, hack=False):
        return self.red_formula.KahanScheme(acc, fout, self.tmp_acc)



from .Abs import Abs
from .Acos import Acos
from .Add import Add
from .ArgMax import ArgMax
from .ArgMin import ArgMin
from .Asin import Asin
from .Atan import Atan
from .Atan2 import Atan2
from .BSpline import BSpline
from .Clamp import Clamp
from .ClampInt import ClampInt
from .Concat import Concat
from .Cos import Cos
from .DiffClampInt import DiffClampInt
from .Divide import Divide
from .Elem import Elem
from .ElemT import ElemT
from .Exp import Exp
from .Extract import Extract
from .ExtractT import ExtractT
from .Floor import Floor
from .GradMatrix import GradMatrix
from .IfElse import IfElse
from .IntInv import IntInv
from .Inv import Inv
from .Log import Log
from .MatVecMult import MatVecMult
from .Max import Max
from .Min import Min
from .Minus import Minus
from .Mod import Mod
from .Mult import Mult
from .Norm2 import Norm2
from .Normalize import Normalize
from .Pow import Pow
from .Powf import Powf
from .ReLU import ReLU
from .Round import Round
from .Rsqrt import Rsqrt
from .Scalprod import Scalprod
from .Sign import Sign
from .Sin import Sin
from .SinXDivX import SinXDivX
from .SqDist import SqDist
from .SqNorm2 import SqNorm2
from .SqNormDiag import SqNormDiag
from .SqNormIso import SqNormIso
from .Sqrt import Sqrt
from .Square import Square
from .Step import Step
from .Subtract import Subtract
from .Sum import Sum
from .SumT import SumT
from .SymSqNorm import SymSqNorm
from .TensorDot import TensorDot
from .TensorProd import TensorProd
from .VecMatMult import VecMatMult
from .WeightedSqDist import WeightedSqDist
from .WeightedSqNorm import WeightedSqNorm
from .XLogX import XLogX

__all__ = [
    "Abs",
    "Acos",
    "Add",
    "ArgMax",
    "ArgMin",
    "Asin",
    "Atan",
    "Atan2",
    "BSpline",
    "Clamp",
    "ClampInt",
    "Concat",
    "Cos",
    "DiffClampInt",
    "Divide",
    "Elem",
    "ElemT",
    "Exp",
    "Extract",
    "ExtractT",
    "GradMatrix",
    "Floor",
    "IfElse",
    "IntInv",
    "Inv",
    "Log",
    "MatVecMult",
    "Max",
    "Min",
    "Minus",
    "Mod",
    "Mult",
    "Norm2",
    "Normalize",
    "Pow",
    "Powf",
    "ReLU",
    "Round",
    "Rsqrt",
    "Scalprod",
    "Sign",
    "Sin",
    "SinXDivX",
    "SqDist",
    "SqNorm2",
    "SqNormDiag",
    "SqNormIso",
    "Sqrt",
    "Square",
    "Step",
    "Subtract",
    "Sum",
    "SumT",
    "SymSqNorm",
    "TensorDot",
    "TensorProd",
    "VecMatMult",
    "WeightedSqDist",
    "WeightedSqNorm",
    "XLogX",
]



from keopscore.formulas.Operation import Broadcast
from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.formulas.maths.Scalprod import Scalprod
from keopscore.formulas.maths.Sum import Sum
from keopscore.formulas.maths.Square import Square
from keopscore.formulas.variables.Zero import Zero
from keopscore.utils.misc_utils import KeOps_Error

##########################
######    Divide     #####
##########################


class Divide_Impl(VectorizedScalarOp):
    """the binary divide operation"""

    string_id = "Divide"
    print_spec = "/", "mid", 3

    def ScalarOp(self, out, arg0, arg1):
        """returns the atomic piece of c++ code to evaluate the function on arg and return the result in out"""
        return f"{out.id} = {arg0.id} / {arg1.id};\n"

    #  \diff_V (A/B) = ((\diff_V A) * B - A * (\diff_V B)) / B^2
    def DiffT(self, v, gradin):
        fa, fb = self.children
        if fa.dim == 1 and fb.dim > 1:
            return fa.DiffT(v, Sum(gradin / fb)) - fb.DiffT(v, fa * gradin / Square(fb))
        elif fb.dim == 1 and fa.dim > 1:
            return (
                fa.DiffT(v, fb * gradin) - fb.DiffT(v, Scalprod(gradin, fa))
            ) / Square(fb)
        else:
            return fa.DiffT(v, gradin / fb) - fb.DiffT(v, fa * gradin / Square(fb))

    # parameters for testing the operation (optional)
    nargs = 2  # number of arguments


# N.B. The following separate function should theoretically be implemented
# as a __new__ method of the previous class, but this can generate infinite recursion problems
def Divide(arg0, arg1):
    if isinstance(arg0, Zero):
        return Broadcast(arg0, arg1.dim)
    elif isinstance(arg1, Zero):
        KeOps_Error("division by zero")
    else:
        return Divide_Impl(arg0, arg1)



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.utils.math_functions import keops_diffclampint

"""
//////////////////////////////////////////////////////////////
////         DIFFCLAMPINT : DiffClampInt< F, A, B >       ////
//////////////////////////////////////////////////////////////

// DiffClampInt(x,a,b) = 0 if x<a, 1 if a<=x<=b, 0 if b<x 
// N.B. used as derivative of ClampInt operation
"""


class DiffClampInt(VectorizedScalarOp):

    string_id = "DiffClampInt"

    def __init__(self, x, a, b):
        super().__init__(x, params=(a, b))

    ScalarOpFun = keops_diffclampint

    def DiffT(self, v, gradin):
        from keopscore.formulas import Zero

        return Zero(v.dim)

    # parameters for testing the operation (optional)
    enable_test = (
        False  # (because it will be tested anyway if we test the gradient of ClampInt)
    )



from keopscore.formulas.Operation import Operation
from keopscore.utils.code_gen_utils import (
    c_variable,
    c_for_loop,
)


####################################
######    Tensor product       #####
####################################


class TensorProd(Operation):
    string_id = "TensorProd"

    def __init__(self, arg0, arg1):
        super().__init__(arg0, arg1)
        self.dim = arg0.dim * arg1.dim

    def Op(self, out, table, arg0, arg1):
        q = c_variable("int")
        loop, k = c_for_loop(0, arg0.dim, 1, pragma_unroll=True)
        inner_loop, l = c_for_loop(0, arg1.dim, 1, pragma_unroll=True)
        return f"""
                    #if C_CONTIGUOUS     // row major
                        {q.declare_assign(0)}
                        {loop(inner_loop(out[q].assign(arg0[k] * arg1[l]) + q.add_assign(1)))}
                    #else               // column major
                        {q.declare_assign(0)}
                        {loop(inner_loop(out[k + l * arg0.dim].assign(arg0[k] * arg1[l]) + q.add_assign(1)))}
                    #endif
                """

    def DiffT(self, v, gradin):
        from keopscore.formulas import MatVecMult, VecMatMult

        f, g = self.children
        return f.DiffT(v, MatVecMult(gradin, g)) + g.DiffT(v, VecMatMult(f, gradin))



from .ComplexAbs import ComplexAbs
from .ComplexAdd import ComplexAdd
from .ComplexAngle import ComplexAngle
from .ComplexDivide import ComplexDivide
from .ComplexExp import ComplexExp
from .ComplexExp1j import ComplexExp1j
from .ComplexImag import ComplexImag
from .ComplexMult import ComplexMult
from .ComplexReal import ComplexReal
from .ComplexRealScal import ComplexRealScal
from .ComplexSquareAbs import ComplexSquareAbs
from .ComplexSubtract import ComplexSubtract
from .ComplexSum import ComplexSum
from .ComplexSumT import ComplexSumT
from .Conj import Conj
from .Imag2Complex import Imag2Complex
from .Real2Complex import Real2Complex



from keopscore.formulas.reductions.KMin_ArgKMin_Reduction import KMin_ArgKMin_Reduction
from keopscore.formulas.reductions.Zero_Reduction import Zero_Reduction
from keopscore.utils.code_gen_utils import (
    c_for_loop,
    new_c_varname,
    c_variable,
)


class ArgKMin_Reduction(KMin_ArgKMin_Reduction):
    """Implements the arg-k-min reduction operation : for each i or each j, find the indices of the
    k minimal values of Fij operation is vectorized: if Fij is vector-valued, arg-k-min is computed
    for each dimension."""

    string_id = "ArgKMin_Reduction"

    def __init__(self, formula, K, tagIJ):
        super().__init__(formula, K, tagIJ)
        self.dim = K * formula.dim

    def FinalizeOutput(self, acc, out, i):
        fdim = self.formula.dim
        p = c_variable("int", new_c_varname("p"))
        loop, k = c_for_loop(0, fdim, 1, pragma_unroll=True)
        body = p.declare_assign(k)
        inner_loop, l = c_for_loop(
            k, k + 2 * self.K * fdim, 2 * fdim, pragma_unroll=True
        )
        body += inner_loop(out[p].assign(acc[l + fdim]) + p.add_assign(fdim))
        return loop(body)
        outer_body

    def DiffT(self, v, gradin):
        return Zero_Reduction(v.dim, v.cat % 2)



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.utils.math_functions import keops_rcp


##########################
######    INVERSE : Inv<F>        #####
##########################


class Inv(VectorizedScalarOp):
    """the "Inv" vectorized operation"""

    string_id = "Inv"
    print_spec = "1/", "pre", 2

    ScalarOpFun = keops_rcp

    @staticmethod
    def Derivative(f):
        return -1 / f**2

    # parameters for testing the operation (optional)
    torch_op = "lambda x : 1 / x"



from keopscore.formulas.VectorizedComplexScalarOp import VectorizedComplexScalarOp
from keopscore.utils.code_gen_utils import (
    c_for_loop,
    new_c_varname,
    c_variable,
)
from keopscore.utils.math_functions import keops_exp, keops_cos, keops_sin
from keopscore.utils.code_gen_utils import c_variable
from keopscore.formulas.complex.ComplexReal import ComplexReal
from keopscore.formulas.complex.ComplexImag import ComplexImag
from keopscore.formulas.complex.Real2Complex import Real2Complex
from keopscore.formulas.complex.Imag2Complex import Imag2Complex
from keopscore.formulas.maths.Exp import Exp
from keopscore.formulas.maths.Cos import Cos
from keopscore.formulas.maths.Sin import Sin


# /////////////////////////////////////////////////////////////////////////
# ////      ComplexExp                           ////
# /////////////////////////////////////////////////////////////////////////


class ComplexExp(VectorizedComplexScalarOp):

    string_id = "ComplexExp"

    def ScalarOp(self, out, inF):
        r = c_variable(out.dtype, new_c_varname("r"))
        string = r.declare_assign(keops_exp(inF[0]))
        string += out[0].assign(r * keops_cos(inF[1]))
        string += out[1].assign(r * keops_sin(inF[1]))
        return string

    # building equivalent formula for autodiff
    def DiffT(self, v, gradin):
        f = self.children[0]
        AltAbs = Exp(ComplexReal(f))
        AltReal = AltAbs * Cos(ComplexImag(f))
        AltImag = AltAbs * Sin(ComplexImag(f))
        AltFormula = Real2Complex(AltReal) + Imag2Complex(AltImag)
        return AltFormula.DiffT(v, gradin)



from keopscore.formulas.maths.Rsqrt import Rsqrt
from keopscore.formulas.maths.SqNorm2 import SqNorm2


class Normalize:
    def __new__(cls, arg):
        return Rsqrt(SqNorm2(arg)) * arg

    enable_test = False



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.formulas.variables.Zero import Zero


##########################
######    Minus      #####
##########################


class Minus_Impl(VectorizedScalarOp):
    """the "minus" vectorized operation"""

    string_id = "Minus"
    print_spec = "-", "pre", 2

    def ScalarOp(self, out, arg):
        # returns the atomic piece of c++ code to evaluate the function on arg and return
        # the result in out
        return f"{out.id} = -{arg.id};\n"

    def DiffT(self, v, gradin):
        f = self.children[0]
        return -f.DiffT(v, gradin)

    # parameters for testing the operation (optional)
    nargs = 1  # number of arguments
    torch_op = "torch.neg"  # equivalent PyTorch operation


# N.B. The following separate function should theoretically be implemented
# as a __new__ method of the previous class, but this can generate infinite recursion problems
def Minus(arg):
    if isinstance(arg, Zero):
        return arg
    else:
        return Minus_Impl(arg)



from keopscore.utils.code_gen_utils import VectCopy
from keopscore.formulas.reductions.Max_ArgMax_Reduction_Base import (
    Max_ArgMax_Reduction_Base,
)


class Max_ArgMax_Reduction(Max_ArgMax_Reduction_Base):
    """Implements the max+argmax reduction operation : for each i or each j, find the maximal value of Fij and its index
    operation is vectorized: if Fij is vector-valued, max+argmax is computed for each dimension."""

    string_id = "Max_ArgMax_Reduction"

    def __init__(self, formula, tagIJ):
        super().__init__(formula, tagIJ)
        self.dim = 2 * formula.dim

    def FinalizeOutput(self, acc, out, i):
        return VectCopy(out, acc)



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.formulas.variables.Zero import Zero
from keopscore.utils.math_functions import keops_sign


##########################
######    Sign        #####
##########################


class Sign(VectorizedScalarOp):
    """the sign vectorized operation"""

    string_id = "Sign"

    ScalarOpFun = keops_sign

    def DiffT(self, v, gradin):
        return Zero(v.dim)



from keopscore.formulas.Operation import Operation
from keopscore.utils.code_gen_utils import c_zero_float, VectCopy
from keopscore.utils.misc_utils import KeOps_Error

# //////////////////////////////////////////////////////////////
# ////     VECTOR "INJECTION" : ExtractT<F,START,DIM>       ////
# //////////////////////////////////////////////////////////////


class ExtractT(Operation):
    string_id = "ExtractT"

    def __init__(self, f, start, dim):
        if start + f.dim > dim or start < 0:
            KeOps_Error("Index out of bound in ExtractT")
        super().__init__(f, params=(start, dim))
        self.start = start
        self.dim = dim
        self.dimarg = f.dim

    def Op(self, out, table, arg):
        # returns the atomic piece of c++ code to evaluate the function on arg and return
        # the result in out
        out_prev, out_mid, out_end = out.split(
            self.start, self.dimarg, self.dim - self.start - self.dimarg
        )
        return "\n".join(
            (
                out_prev.assign(c_zero_float),
                VectCopy(out_mid, arg),
                out_end.assign(c_zero_float),
            )
        )

    def DiffT(self, v, gradin):
        from keopscore.formulas.maths.Extract import Extract

        f = self.children[0]
        return f.DiffT(v, Extract(gradin, self.start, f.dim))

    # parameters for testing the operation (optional)
    enable_test = True  # enable testing for this operation
    nargs = 1  # number of arguments
    test_argdims = [5]  # dimensions of arguments for testing
    test_params = [3, 10]  # dimensions of parameters for testing

    def torch_op(x, s, d):  # equivalent PyTorch operation
        import torch

        out = torch.zeros((*x.shape[:-1], d), device=x.device, dtype=x.dtype)
        out[..., s : (s + x.shape[-1])] = x
        return out



from keopscore.formulas.Operation import Operation
from keopscore.formulas.maths.ArgMin import ArgMin
from keopscore.formulas.maths.OneHot import OneHot
from keopscore.utils.code_gen_utils import c_for_loop, c_if, value
from keopscore.utils.misc_utils import KeOps_Error

############################
######    Min       #####
############################


class Min(Operation):
    string_id = "Min"

    def __init__(self, f):
        super().__init__(f)
        if f.dim < 1:
            KeOps_Error("Min operation is only possible when dimension is non zero.")
        self.dim = 1

    def Op(self, out, table, arg):
        f = self.children[0]
        loop, k = c_for_loop(1, f.dim, 1, pragma_unroll=True)
        string = value(out).assign(arg[0])
        if out.dtype == "half2":
            loop_string = f"""
                // we have to work element-wise...
                __half2 cond = __hgt2(*{out.id},{arg[k].id});                       // cond = (out > outF[k]) (element-wise)
                __half2 negcond = __float2half2_rn(1.0f)-cond;                      // negcond = 1-cond
                *{out.id} = cond * {arg[k].id} + negcond * *{out.id};               // out  = cond * outF[k] + (1-cond) * out
                            """
            string += loop(loop_string)
        else:
            string += loop(c_if(arg[k] < value(out), value(out).assign(arg[k])))
        return string

    def DiffT(self, v, gradin):
        f = self.children[0]
        return f.DiffT(v, OneHot(ArgMin(f), f.dim) * gradin)

    # parameters for testing the operation (optional)
    enable_test = True  # enable testing for this operation
    nargs = 1  # number of arguments
    test_argdims = [5]  # dimensions of arguments for testing
    torch_op = "lambda x : torch.min(x, dim=-1, keepdim=True)[0].type(x.dtype)"



import os
from hashlib import sha256

from keopscore.config.config import disable_pragma_unrolls
from keopscore.utils.misc_utils import KeOps_Error, KeOps_Message


def get_hash_name(*args):
    return sha256("".join(list(str(arg) for arg in args)).encode("utf-8")).hexdigest()[
        :10
    ]


#######################################################################
# .  Python to C++ meta programming toolbox
#######################################################################


def sizeof(dtype):
    if dtype == "float":
        return 4
    elif dtype == "double":
        return 8
    elif dtype == "half":
        return 2
    else:
        KeOps_Error("not implemented")


class new_c_varname:
    # class to generate unique names for variables in C++ code, to avoid conflicts
    dict_instances = {}

    def __new__(self, template_string_id, num=1, as_list=False):
        # - template_string_id is a string, the base name for c_variable
        # - if num>1 returns a list of num new names with same base names
        # For example the first call to new_c_variable("x")
        # will return "x_1", the second call will return "x_2", etc.
        if num > 1 or as_list:
            return list(new_c_varname(template_string_id) for k in range(num))
        if template_string_id in new_c_varname.dict_instances:
            cnt = new_c_varname.dict_instances[template_string_id] + 1
        else:
            cnt = 0
        new_c_varname.dict_instances[template_string_id] = cnt
        string_id = template_string_id + "_" + str(cnt)
        return string_id


class c_variable:
    # class to represent a C++ variable, storing its c++ name and its C++ type.
    def __new__(self, dtype, list_string_id=None):
        if isinstance(list_string_id, list):
            return list(c_variable(dtype, string_id) for string_id in list_string_id)
        else:
            return super(c_variable, self).__new__(self)

    def __init__(self, dtype, string_id=None):
        if string_id is None:
            string_id = new_c_varname("var")
        self.dtype = dtype  # dtype is C++ type of variable
        self.id = string_id  # string_id is C++ name of variable

    def __repr__(self):
        # method for printing the c_variable inside Python code
        return self.id

    def declare(self):
        return f"{self.dtype} {self.id};\n"

    def declare_assign(self, value):
        return f"{self.dtype} " + self.assign(value)

    def assign(self, value):
        if type(value) in (int, float):
            dtype = "int" if type(value) == int else "float"
            return self.assign(c_variable(dtype, str(value)))
        elif type(value) == str:
            return f"{self.id} = ({self.dtype})({value});\n"
        elif value.dtype != self.dtype:
            if self.dtype == "float2" and value.dtype == "float":
                return f"""
                            {self.id}.x = {value.id};
                            {self.id}.y = {value.id};
                        """
            else:
                return f"{self.id} = {cast_to(self.dtype, value)};\n"
        else:
            return f"{self.id} = ({value.id});\n"

    def add_assign(self, value):
        if type(value) in (int, float):
            dtype = "int" if type(value) == int else "float"
            return self.add_assign(c_variable(dtype, str(value)))
        if type(value) == str:
            return f"{self.id} += ({self.dtype})({value});\n"
        elif value.dtype != self.dtype:
            if self.dtype == "float2" and value.dtype == "half2":
                return f"""
                            {self.id}.x += (float){value.id}.x;
                            {self.id}.y += (float){value.id}.y;
                        """
            else:
                return f"{self.id} += {cast_to(self.dtype, value)};\n"
        else:
            return f"{self.id} += ({value.id});\n"

    def binary_op(self, other, python_op, c_op, name, dtype=None):
        if type(other) in (int, float):
            dtype = "int" if type(other) == int else "float"
            return python_op(self, c_variable(dtype, str(other)))
        elif type(other) == c_variable:
            if self.dtype != other.dtype:
                KeOps_Error(
                    f"{name} of two c_variable is only possible with same dtype"
                )
            if dtype is None:
                dtype = self.dtype
            return c_variable(dtype, f"({self.id}{c_op}{other.id})")
        else:
            KeOps_Error("not implemented")

    def __add__(self, other):
        python_op = lambda x, y: x + y
        return self.binary_op(other, python_op, "+", "addition")

    def __mul__(self, other):
        python_op = lambda x, y: x * y
        return self.binary_op(other, python_op, "*", "product")

    def __sub__(self, other):
        python_op = lambda x, y: x - y
        return self.binary_op(other, python_op, "-", "subtraction")

    def __truediv__(self, other):
        python_op = lambda x, y: x / y
        return self.binary_op(other, python_op, "/", "division")

    def __lt__(self, other):
        python_op = lambda x, y: x < y
        return self.binary_op(other, python_op, "<", "comparison", dtype="bool")

    def __le__(self, other):
        python_op = lambda x, y: x <= y
        return self.binary_op(other, python_op, "<=", "comparison", dtype="bool")

    def __gt__(self, other):
        python_op = lambda x, y: x > y
        return self.binary_op(other, python_op, ">", "comparison", dtype="bool")

    def __ge__(self, other):
        python_op = lambda x, y: x >= y
        return self.binary_op(other, python_op, ">=", "comparison", dtype="bool")

    # N.B.: The & symbol (__and__) denotes the bitwise operator, not the logical one
    def logical_and(self, other):
        python_op = lambda x, y: x and y
        return self.binary_op(other, python_op, "&&", "comparison", dtype="bool")

    # N.B.: The | symbol (__or__) denotes the bitwise operator, not the logical one
    def logical_or(self, other):
        python_op = lambda x, y: x or y
        return self.binary_op(other, python_op, "||", "comparison", dtype="bool")

    def ternary(self, out_true, out_false):
        if self.dtype != "bool":
            KeOps_Error(
                f"The input of a ternary operator should have bool dtype, "
                f"found {self.dtype}."
            )
        if out_true.dtype != out_false.dtype:
            KeOps_Error(
                f"The two possible output of a ternary operator should have the same dtype, "
                f"found {out_true.dtype} and {out_false.dtype}."
            )
        return c_variable(
            out_true.dtype, f"({self.id} ? {out_true.id} : {out_false.id})"
        )

    def __neg__(self):
        return c_variable(self.dtype, f"(-{self.id})")

    def __getitem__(self, other):
        if type(other) == int:
            return self[c_variable("int", str(other))]
        elif type(other) == c_variable:
            if other.dtype != "int":
                KeOps_Error("v[i] with i and v c_variable requires i.dtype='int' ")
            return c_variable(value(self.dtype), f"{self.id}[{other.id}]")
        else:
            KeOps_Error("not implemented")


def use_pragma_unroll(n=64):
    if disable_pragma_unrolls:
        return "\n"
    else:
        if n is None:
            return f"\n#pragma unroll\n"
        else:
            return f"\n#pragma unroll({n})\n"


def c_for_loop(start, end, incr, pragma_unroll=False):
    def to_string(x):
        if type(x) == c_variable:
            if x.dtype != "int":
                KeOps_Error("only simple int type for loops implemented")
            return x.id
        elif type(x) == int:
            return str(x)
        else:
            KeOps_Error("only simple int type for loops implemented")

    start, end, incr = map(to_string, (start, end, incr))
    k = c_variable("int", new_c_varname("k"))

    def printfun(body_code):
        string = ""
        if pragma_unroll:
            string += use_pragma_unroll()
        string += f""" for(int {k.id}={start}; {k.id}<{end}; {k.id}+=({incr})) {{
                            {body_code}
                        }}
                    """
        return string

    return printfun, k


c_zero_int = c_variable("int", "0")
c_zero_float = c_variable("float", "0.0f")


def neg_infinity(dtype):
    return c_variable(dtype, f"-({infinity(dtype).id})")


def infinity(dtype):
    if dtype == "float":
        code = "( 1.0f/0.0f )"
    elif dtype == "double":
        code = "( 1.0/0.0 )"
    else:
        KeOps_Error(
            "only float and double dtypes are implemented in new python engine for now"
        )
    return c_variable(dtype, code)


def cast_to(dtype, var):
    # returns C++ code string to do a cast ; e.g. "(float)" if dtype is "float" for example
    simple_dtypes = ["float", "double", "int", "bool"]
    if (dtype in simple_dtypes) and (var.dtype in simple_dtypes):
        return f"({dtype})({var.id})"
    elif dtype == "half2" and var.dtype == "float":
        return f"__float2half2_rn({var.id})"
    elif dtype == "float2" and var.dtype == "half2":
        return f"__half22float2({var.id})"
    elif dtype == "half2" and var.dtype == "float2":
        return f"__float22half2_rn({var.id})"
    else:
        KeOps_Error(f"not implemented: casting from {var.dtype} to {dtype}")


def value(x):
    # either convert c_array or c_variable representing a pointer to its value c_variable (dereference)
    # or converts string "dtype*" to "dtype"
    if isinstance(x, c_array):
        return c_variable(x.dtype, f"(*{x.id})")
    if isinstance(x, c_variable):
        return c_variable(value(x.dtype), f"(*{x.id})")
    elif isinstance(x, str):
        if x[-1] == "*":
            return x[:-1]
        else:
            KeOps_Error(
                "Incorrect input string in value function; it should represent a pointer C++ type."
            )
    else:
        KeOps_Error("input should be either c_variable instance or string.")


def pointer(x):
    # either convert c_variable to its address c_variable (reference)
    # or converts string "dtype" to "dtype*"
    if isinstance(x, c_variable):
        return c_variable(pointer(x.dtype), f"(&{x.id})")
    elif isinstance(x, str):
        return x + "*"
    else:
        KeOps_Error("input should be either c_variable instance or string.")


class c_array:
    def __init__(self, dtype, dim, string_id=new_c_varname("array")):
        if dim < 0:
            KeOps_Error("negative dimension for array")
        self.c_var = c_variable(pointer(dtype), string_id)
        self.dtype = dtype
        self.dim = dim
        self.id = string_id

    def __repr__(self):
        # method for printing the c_variable inside Python code
        return self.c_var.__repr__()

    def declare(self):
        # returns C++ code to declare a fixed-size arry of size dim,
        # skipping declaration if dim=0
        if self.dim > 0:
            return f"{self.dtype} {self.c_var.id}[{self.dim}];"
        else:
            return ""

    def split(self, *dims):
        # split c_array in n sub arrays with dimensions dims[0], dims[1], ..., dims[n-1]
        if sum(dims) != self.dim:
            KeOps_Error("incompatible dimensions for split")
        listarr, cumdim = [], 0
        for dim in dims:
            listarr.append(c_array(self.dtype, dim, f"({self.id}+{cumdim})"))
            cumdim += dim
        return listarr

    def assign(self, val):
        # returns C++ code string to fill all elements of a fixed size array with a single value
        # val is a c_variable representing the value.
        loop, k = c_for_loop(0, self.dim, 1)
        return loop(self[k].assign(val))

    def __getitem__(self, other):
        if type(other) == int:
            return self[c_variable("int", str(other))]
        elif type(other) == c_variable:
            if other.dtype != "int":
                KeOps_Error("v[i] with i and v c_array requires i.dtype='int' ")
            return c_variable(self.dtype, f"{self.id}[{other.id}]")
        else:
            KeOps_Error("not implemented")

    @property
    def c_print(self):
        if self.dtype in ["float", "double"]:
            tag = "%f, " * self.dim
        elif self.dtype in ["int", "float*", "double*"]:
            tag = "%d, " * self.dim
        else:
            KeOps_Error(f"c_print not implemented for dtype={self.dtype}")
        string = f'printf("{self.id} = {tag}\\n"'
        for i in range(self.dim):
            string += f", {self[i].id}"
        string += ");\n"
        return string


def VectApply(fun, out, *args):
    # returns C++ code string to apply a scalar operation to fixed-size arrays, following broadcasting rules.
    # - fun is the scalar unary function to be applied, it must accept two c_variable or c_array inputs and output a string
    # - out must be a c_array instance
    # - args may be c_array or c_variable instances
    #
    # Example : if out.dim = 3, arg0.dim = 1, arg1.dim = 3,
    # it will generate the following (in pseudo-code for clarity) :
    #   #pragma unroll
    #   for(int k=0; k<out.dim; k++)
    #       fun(out[k], arg0[0], arg1[k]);
    #
    # Equivalently, if out.dim = 3, arg0 is c_variable, arg1.dim = 3,
    # it will generate the following (in pseudo-code for clarity) :
    #   #pragma unroll
    #   for(int k=0; k<out.dim; k++)
    #       fun(out[k], arg0, arg1[k]);

    dims = [out.dim]
    for arg in args:
        if isinstance(arg, c_variable):
            dims.append(1)
        elif isinstance(arg, c_array):
            dims.append(arg.dim)
        else:
            KeOps_Error("args must be c_variable or c_array instances")
    dimloop = max(dims)
    if not set(dims) in ({dimloop}, {1, dimloop}):
        KeOps_Error("incompatible dimensions in VectApply")
    incr_out = 1 if out.dim == dimloop else 0
    incr_args = list((1 if dim == dimloop else 0) for dim in dims[1:])

    forloop, k = c_for_loop(0, dimloop, 1, pragma_unroll=True)

    argsk = []
    for (arg, incr) in zip(args, incr_args):
        if isinstance(arg, c_variable):
            argsk.append(arg)
        elif isinstance(arg, c_array):
            argsk.append(arg[k * incr])

    return forloop(fun(out[k * incr_out], *argsk))


def ComplexVectApply(fun, out, *args):
    # similar to VectApply but for complex operations

    dims = [out.dim]
    for arg in args:
        if isinstance(arg, c_array):
            dims.append(arg.dim)
        else:
            KeOps_Error("args must be c_array instances")
    dimloop = max(dims)
    if not set(dims) in ({dimloop}, {2, dimloop}):
        KeOps_Error("incompatible dimensions in ComplexVectApply")
    incr_out = 1 if out.dim == dimloop else 0
    incr_args = list((1 if dim == dimloop else 0) for dim in dims[1:])

    forloop, k = c_for_loop(0, dimloop, 2, pragma_unroll=True)

    argsk = []
    for (arg, incr) in zip(args, incr_args):
        argk = c_array(arg.dtype, 2, f"({arg.id}+{k.id}*{incr})")
        argsk.append(argk)
    outk = c_array(out.dtype, 2, f"({out.id}+{k.id}*{incr_out})")
    return forloop(fun(outk, *argsk))


def VectCopy(out, arg, dim=None):
    # returns a C++ code string representing a vector copy between fixed-size arrays
    # - dim is dimension of arrays
    # - out is c_variable representing the output array
    # - arg is c_variable representing the input array
    if dim is None:
        dim = out.dim
    forloop, k = c_for_loop(0, dim, 1, pragma_unroll=True)
    return forloop(out[k].assign(arg[k]))


def call_list(args):
    return ", ".join(list(arg.id for arg in args))


def signature_list(args):
    return ", ".join(list(f"{arg.dtype} {arg.id}" for arg in args))


def c_include(*headers):
    return "".join(f"#include <{header}>\n" for header in headers)


def c_if(condition, command, else_command=None):
    string = f""" if ({condition.id}) {{
                      {command}
                }}
            """
    if else_command:
        string += f""" else {{
                      {else_command}
                }}
            """
    return string


def c_block(*commands):
    block_string = "".join(commands)
    return f""" {{
                      {block_string}
                }}
            """


def c_function(name, dtypeout, args, commands, qualifier=None):
    # first write the signature of the function :
    string = ""
    if qualifier is not None:
        string += f"{qualifier} "
    string += f"{dtypeout} {name}({signature_list(args)}) "
    # then the body
    string += "\n{\n"
    string += "\n".join(list(c for c in commands))
    string += "\n}\n"
    return string


#######################################################################
# .  KeOps related helpers
#######################################################################


def GetDims(Vars):
    # returns the list of dim fields (dimensions) of a list of Var instances
    return tuple(v.dim for v in Vars)


def GetInds(Vars):
    # returns the list of ind fields (indices) of a list of Var instances
    return tuple(v.ind for v in Vars)


class Var_loader:
    def __init__(self, red_formula):

        formula = red_formula.formula
        tagI, tagJ = red_formula.tagI, red_formula.tagJ

        mymin = lambda x: min(x) if len(x) > 0 else -1

        self.Varsi = formula.Vars(
            cat=tagI
        )  # list all "i"-indexed variables in the formula
        self.nvarsi = len(self.Varsi)  # number of "i"-indexed variables
        self.indsi = GetInds(self.Varsi)  # list indices of "i"-indexed variables
        self.pos_first_argI = mymin(self.indsi)  # first index of "i"-indexed variables
        self.dimsx = GetDims(self.Varsi)  # list dimensions of "i"-indexed variables
        self.dimx = sum(self.dimsx)  # total dimension of "i"-indexed variables

        self.Varsj = formula.Vars(
            cat=tagJ
        )  # list all "j"-indexed variables in the formula
        self.nvarsj = len(self.Varsj)  # number of "j"-indexed variables
        self.indsj = GetInds(self.Varsj)  # list indices of "j"-indexed variables
        self.pos_first_argJ = mymin(self.indsj)  # first index of "j"-indexed variables
        self.dimsy = GetDims(self.Varsj)  # list dimensions of "j"-indexed variables
        self.dimy = sum(self.dimsy)  # total dimension of "j"-indexed variables

        self.Varsp = formula.Vars(cat=2)  # list all parameter variables in the formula
        self.nvarsp = len(self.Varsp)  # number of parameter variables
        self.indsp = GetInds(self.Varsp)  # list indices of parameter variables
        self.pos_first_argP = mymin(self.indsp)  # first index of parameter variables
        self.dimsp = GetDims(self.Varsp)  # list indices of parameter variables
        self.dimp = sum(self.dimsp)  # total dimension of parameter variables

        self.inds = GetInds(formula.Vars_)
        self.nminargs = max(self.inds) + 1 if len(self.inds) > 0 else 0

    def table(self, xi, yj, pp):
        return table(
            self.nminargs,
            self.dimsx,
            self.dimsy,
            self.dimsp,
            self.indsi,
            self.indsj,
            self.indsp,
            xi,
            yj,
            pp,
        )

    def direct_table(self, args, i, j):
        return direct_table(
            self.nminargs,
            self.dimsx,
            self.dimsy,
            self.dimsp,
            self.indsi,
            self.indsj,
            self.indsp,
            args,
            i,
            j,
        )

    def load_vars(self, cat, *args, **kwargs):
        if cat == "i":
            dims, inds = self.dimsx, self.indsi
        elif cat == "j":
            dims, inds = self.dimsy, self.indsj
        elif cat == "p":
            dims, inds = self.dimsp, self.indsp
        return load_vars(dims, inds, *args, **kwargs)


def table(nminargs, dimsx, dimsy, dimsp, indsi, indsj, indsp, xi, yj, pp):
    res = [None] * nminargs
    for (dims, inds, xloc) in (
        (dimsx, indsi, xi),
        (dimsy, indsj, yj),
        (dimsp, indsp, pp),
    ):
        k = 0
        for u in range(len(dims)):
            res[inds[u]] = c_array(xloc.dtype, dims[u], f"({xloc.id}+{k})")
            k += dims[u]
    return res


def direct_table(nminargs, dimsx, dimsy, dimsp, indsi, indsj, indsp, args, i, j):
    res = [None] * nminargs
    for (dims, inds, row_index) in (
        (dimsx, indsi, i),
        (dimsy, indsj, j),
        (dimsp, indsp, c_zero_int),
    ):
        for u in range(len(dims)):
            arg = args[inds[u]]
            res[inds[u]] = c_array(
                value(arg.dtype), dims[u], f"({arg.id}+{row_index.id}*{dims[u]})"
            )
    return res


def table4(
    nminargs,
    dimsx,
    dimsy,
    dimsp,
    dims_new,
    indsi,
    indsj,
    indsp,
    inds_new,
    xi,
    yj,
    pp,
    arg_new,
):
    res = [None] * nminargs
    for (dims, inds, xloc) in (
        (dimsx, indsi, xi),
        (dimsy, indsj, yj),
        (dimsp, indsp, pp),
        (dims_new, inds_new, arg_new),
    ):
        k = 0
        for u in range(len(dims)):
            res[inds[u]] = c_array(xloc.dtype, dims[u], f"({xloc.id}+{k})")
            k += dims[u]

    return res


def load_vars(dims, inds, xloc, args, row_index=c_zero_int, offsets=None, indsref=None):
    # returns a c++ code used to create a local copy of slices of the input tensors, for evaluating a formula
    # - dims is a list of integers giving dimensions of variables
    # - dims is a list of integers giving indices of variables
    # - xloc is a c_array, the local array which will receive the copy
    # - args is a list of c_variable, representing pointers to input tensors
    # - row_index is a c_variable (of dtype="int"), specifying which row of the matrix should be loaded
    # - offsets is an optional c_array (of dtype="int"), specifying variable-dependent offsets (used when broadcasting batch dimensions)
    # - indsref is an optional list of integers, giving index mapping for offsets
    #
    # Example: assuming i=c_variable("int", "5"), xloc=c_variable("float", "xi") and px=c_variable("float**", "px"), then
    # if dims = [2,2,3] and inds = [7,9,8], the call to
    #   load_vars (dims, inds, xi, [arg0, arg1,..., arg9], row_index=i )
    # will output the following code:
    #   xi[0] = arg7[5*2+0];
    #   xi[1] = arg7[5*2+1];
    #   xi[2] = arg9[5*2+0];
    #   xi[3] = arg9[5*2+1];
    #   xi[4] = arg8[5*3+0];
    #   xi[5] = arg8[5*3+1];
    #   xi[6] = arg8[5*3+2];
    #
    # Example (with offsets): assuming i=c_variable("int", "5"),
    # xloc=c_variable("float", "xi"), px=c_variable("float**", "px"),
    # and offsets = c_array("int", 3, "offsets"), then
    # if dims = [2,2,3] and inds = [7,9,8], the call to
    #   load_vars (dims, inds, xi, [arg0, arg1,..., arg9], row_index=i, offsets=offsets)
    # will output the following code:
    #   xi[0] = arg7[(5+offsets[0])*2+0];
    #   xi[1] = arg7[(5+offsets[0])*2+1];
    #   xi[2] = arg9[(5+offsets[1])*2+0];
    #   xi[3] = arg9[(5+offsets[1])*2+1];
    #   xi[4] = arg8[(5+offsets[2])*3+0];
    #   xi[5] = arg8[(5+offsets[2])*3+1];
    #   xi[6] = arg8[(5+offsets[2])*3+2];
    #
    # Example (with offsets and indsref): assuming i=c_variable("int", "5"),
    # xloc=c_variable("float", "xi"), px=c_variable("float**", "px"),
    # offsets = c_array("int", 3, "offsets"),
    # if dims = [2,2,3] and inds = [7,9,8],
    # and indsref = [8,1,7,3,9,2], then since 7,8,9 are at positions 2,0,4 in indsref,
    # the call to
    #   load_vars (dims, inds, xi, [arg0, arg1,..., arg9], row_index=i, offsets=offsets, indsref=indsref)
    # will output the following code:
    #   xi[0] = arg7[(5+offsets[2])*2+0];
    #   xi[1] = arg7[(5+offsets[2])*2+1];
    #   xi[2] = arg9[(5+offsets[0])*2+0];
    #   xi[3] = arg9[(5+offsets[0])*2+1];
    #   xi[4] = arg8[(5+offsets[4])*3+0];
    #   xi[5] = arg8[(5+offsets[4])*3+1];
    #   xi[6] = arg8[(5+offsets[4])*3+2];
    string = ""
    if len(dims) > 0:
        string += "{\n"
        string += "int a=0;\n"
        for u in range(len(dims)):
            l = indsref.index(inds[u]) if indsref else u
            row_index_str = (
                f"({row_index.id}+{offsets.id}[{l}])" if offsets else row_index.id
            )
            string += use_pragma_unroll()
            string += f"for(int v=0; v<{dims[u]}; v++) {{\n"
            string += (
                f"    {xloc.id}[a] = {args[inds[u]].id}[{row_index_str}*{dims[u]}+v];\n"
            )
            string += "     a++;\n"
            string += "}\n"
        string += "}\n"
    return string


def load_vars_chunks(
    inds, dim_chunk, dim_chunk_load, dim_org, xloc, args, k, row_index=c_zero_int
):
    #
    # loads chunks of variables, unrolling dimensions and indices.
    #
    # Example:
    #   load_chunks([7,9,8], 3, 2, 11, xi, px, k, row_index=i)
    # with:
    # xi = c_variable("float", "xi"),
    # px = c_variable("float**", "px")
    # i = c_variable("int","5"),
    # k = c_variable("int","k")
    # means : there are 3 chunks of vectors to load. They are located
    # at positions 7, 9 and 8 in px. Now i=5 and dim_org=11, so we start
    # to load vectors at positions px[7]+5*11, px[9]+5*11, px[8]+5*11.
    # For each, we load the kth chunk, assuming vector is divided
    # into chunks of size 3. And finally, we stop after loading 2 values.
    #
    # So we will execute:
    #   xi[0] = px[7][5*11+k*3];
    #   xi[1] = px[7][5*11+k*3+1];
    #   xi[2] = px[9][5*11+k*3];
    #   xi[3] = px[9][5*11+k*3+1];
    #   xi[4] = px[8][5*11+k*3];
    #   xi[5] = px[8][5*11+k*3+1];
    string = ""
    if len(inds) > 0:
        string += "{"
        string += "int a=0;\n"
        for u in range(len(inds)):
            string += use_pragma_unroll()
            string += f"for(int v=0; v<{dim_chunk_load}; v++) {{\n"
            string += f"    {xloc.id}[a] = {args[inds[u]].id}[{row_index.id}*{dim_org}+{k.id}*{dim_chunk}+v];\n"
            string += "     a++;\n"
            string += "}"
        string += "}"
    return string


def load_vars_chunks_offsets(
    inds,
    indsref,
    dim_chunk,
    dim_chunk_load,
    dim_org,
    xloc,
    args,
    k,
    offsets,
    row_index=c_zero_int,
):
    # Version with variable-dependent offsets (used when broadcasting batch dimensions)
    # indsref gives mapping for offsets indexing
    # Example:
    #   load_vars_chunks_offsets([2,3,1], [8,9,7,3,1,2], 3, 2, 11, xi, px, k, offsets, row_index=i)
    # Since 2,3,1 are at positions 5,3,4 respectively in the list [8,9,7,3,1,2],
    # will output:
    #   xi[0] = px[2][(5+offsets[5])*11+k*3];
    #   xi[1] = px[2][(5+offsets[5])*11+k*3+1];
    #   xi[2] = px[3][(5+offsets[3])*11+k*3];
    #   xi[3] = px[3][(5+offsets[3])*11+k*3+1];
    #   xi[4] = px[1][(5+offsets[4])*11+k*3];
    #   xi[5] = px[1][(5+offsets[4])*11+k*3+1];
    string = ""
    if len(inds) > 0:
        string = "{"
        string += "int a=0;\n"
        for u in range(len(inds)):
            l = indsref.index(inds[u])
            string += use_pragma_unroll()
            string += f"for(int v=0; v<{dim_chunk_load}; v++) {{\n"
            string += f"    {xloc.id}[a] = {args[inds[u]].id}[({row_index.id}+{offsets.id}[{l}])*{dim_org}+{k.id}*{dim_chunk}+v];\n"
            string += "     a++;\n"
            string += "}"
        string += "}"
    return string


def varseq_to_array(vars, vars_ptr_name):
    # returns the C++ code corresponding to storing the values of a sequence of variables
    # into an array.

    dtype = vars[0].dtype
    nvars = len(vars)

    # we check that all variables have the same type
    if not all(var.dtype == dtype for var in vars[1:]):
        KeOps_Error("[KeOps] internal error ; incompatible dtypes in varseq_to_array.")
    string = f"""   {dtype} {vars_ptr_name}[{nvars}];
              """
    for i in range(nvars):
        string += f"""  {vars_ptr_name}[{i}] = {vars[i].id};
                   """
    return string


def clean_keops(recompile_jit_binary=True, verbose=True):
    import keopscore.config.config
    from keopscore.config.config import get_build_folder

    build_path = get_build_folder()
    use_cuda = keopscore.config.config.use_cuda
    if use_cuda:
        from keopscore.config.config import jit_binary
    else:
        jit_binary = None
    for f in os.scandir(build_path):
        if recompile_jit_binary or f.path != jit_binary:
            os.remove(f.path)
    if verbose:
        KeOps_Message(f"{build_path} has been cleaned.")
    from keopscore.get_keops_dll import get_keops_dll

    get_keops_dll.reset()
    if use_cuda and recompile_jit_binary:
        from keopscore.binders.nvrtc.Gpu_link_compile import Gpu_link_compile

        Gpu_link_compile.compile_jit_compile_dll()



from keopscore.formulas.maths.Sqrt import Sqrt
from keopscore.formulas.maths.Scalprod import Scalprod


class Norm2:
    def __new__(cls, arg):
        return Sqrt(Scalprod(arg, arg))

    enable_test = False



from keopscore.formulas.Operation import Operation
from keopscore.utils.code_gen_utils import value, c_zero_float, c_for_loop
from keopscore.utils.misc_utils import KeOps_Error

############################
######    ELEMENT "INJECTION" : ElemT(f,m,n)
############################


class ElemT(Operation):
    string_id = "ElemT"

    def __init__(self, f, n, m):
        super().__init__(f, params=(n, m))
        if f.dim != 1:
            KeOps_Error("Input of ElemT should be a scalar")
        self.dim = n
        self.n = n
        self.m = m

    def Op(self, out, table, arg):
        n, m = self.n, self.m
        loop1, k = c_for_loop(0, m, 1, pragma_unroll=True)
        string = loop1(out[k].assign(c_zero_float))
        string += out[m].assign(value(arg))
        loop2, k = c_for_loop(m + 1, n, 1, pragma_unroll=True)
        string += loop2(out[k].assign(c_zero_float))
        return string

    def DiffT(self, v, gradin):
        from keopscore.formulas.maths.Elem import Elem

        f = self.children[0]
        return f.DiffT(v, Elem(gradin, self.m))

    # parameters for testing the operation (optional)
    enable_test = True  # enable testing for this operation
    nargs = 1  # number of arguments
    test_argdims = [1]  # dimensions of arguments for testing
    test_params = [5, 3]  # dimensions of parameters for testing

    def torch_op(x, n, m):  # equivalent PyTorch operation
        import torch

        out = torch.zeros((*x.shape[:-1], n), device=x.device, dtype=x.dtype)
        out[..., m][..., None] = x
        return out



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.formulas.maths.DiffClampInt import DiffClampInt
from keopscore.utils.math_functions import keops_clampint


class ClampInt(VectorizedScalarOp):
    """ClampInt(x,a,b) = a if x<a, x if a<=x<=b, b if b<x
    N.B. same as Clamp but a and b are fixed integers.
    ClampInt may be faster than Clamp because we avoid the transfer
    of A and B in memory.
    """

    string_id = "ClampInt"

    def __init__(self, x, a, b):
        super().__init__(x, params=(a, b))

    ScalarOpFun = keops_clampint

    @staticmethod
    def Derivative(x, a, b):
        return DiffClampInt(x, a, b)

    # parameters for testing the operation (optional)
    test_params = [0, 1]  # parameters to try
    torch_op = "torch.clamp"  # equivalent PyTorch operation



import ctypes
from ctypes.util import find_library

import keopscore.config.config
from keopscore.utils.misc_utils import (
    KeOps_Error,
    KeOps_Warning,
    find_library_abspath,
    KeOps_OS_Run,
)
from keopscore.config.config import cxx_compiler, get_build_folder
import os
from os.path import join

# Some constants taken from cuda.h
CUDA_SUCCESS = 0
CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_BLOCK = 1
CU_DEVICE_ATTRIBUTE_MAX_SHARED_MEMORY_PER_BLOCK = 8

libcuda_folder = os.path.dirname(find_library_abspath("cuda"))
libnvrtc_folder = os.path.dirname(find_library_abspath("nvrtc"))


def get_cuda_include_path():

    # auto detect location of cuda headers

    # First we look at CUDA_PATH env variable if it is set
    path = os.getenv("CUDA_PATH")
    if path:
        path = join(path, "include")
        if os.path.isfile(join(path, "cuda.h")) and os.path.isfile(
            join(path, "nvrtc.h")
        ):
            return path

    # if not successfull, we try a few standard locations:
    cuda_version = get_cuda_version(out_type="string")
    cuda_paths_to_try_start = [
        join(os.path.sep, "opt", "cuda"),
        join(os.path.sep, "usr", "local", "cuda"),
        join(os.path.sep, "usr", "local", f"cuda-{cuda_version}"),
    ]
    cuda_paths_to_try_end = [
        "include",
        join("targets", "x86_64-linux", "include"),
    ]
    for path_start in cuda_paths_to_try_start:
        for path_end in cuda_paths_to_try_end:
            path = join(path_start, path_end)
            if os.path.isfile(join(path, "cuda.h")) and os.path.isfile(
                join(path, "nvrtc.h")
            ):
                return path

    # if not successfull, we try to infer location from the libs
    cuda_include_path = None
    for libpath in libcuda_folder, libnvrtc_folder:
        for libtag in "lib", "lib64":
            libtag = os.path.sep + libtag + os.path.sep
            if libtag in libpath:
                includetag = os.path.sep + "include" + os.path.sep
                includepath = libpath.replace(libtag, includetag)
                if os.path.isfile(join(includepath, "cuda.h")) and os.path.isfile(
                    join(includepath, "nvrtc.h")
                ):
                    return includepath

    # last try, testing if by any chance the header is already in the default
    # include path of gcc
    path_cudah = get_include_file_abspath("cuda.h")
    if path_cudah:
        path = os.path.dirname(path_cudah)
    if os.path.isfile(join(path, "nvrtc.h")):
        return path

    # finally nothing found, so we display a warning asking the user to do something
    KeOps_Warning(
        """
    The location of Cuda header files cuda.h and nvrtc.h could not be detected on your system.
    You must determine their location and then define the environment variable CUDA_PATH,
    either before launching Python or using os.environ before importing keops. For example
    if these files are in /vol/cuda/10.2.89-cudnn7.6.4.38/include you can do :
      import os
      os.environ['CUDA_PATH'] = '/vol/cuda/10.2.89-cudnn7.6.4.38'
    """
    )


def get_include_file_abspath(filename):
    tmp_file = join(get_build_folder(), "tmp.txt")
    KeOps_OS_Run(
        f'echo "#include <{filename}>" | {cxx_compiler} -M -E -x c++ - | head -n 2 > {tmp_file}'
    )
    strings = open(tmp_file).read().split()
    abspath = None
    for s in strings:
        if filename in s:
            abspath = s
    os.remove(tmp_file)
    return abspath


def cuda_include_fp16_path():
    """
    We look for float 16 cuda headers cuda_fp16.h and cuda_fp16.hpp
    based on cuda_path locations and return their directory
    """
    from keopscore.config.config import cuda_include_path

    if cuda_include_path:
        return cuda_include_path
    cuda_fp16_h_abspath = get_include_file_abspath("cuda_fp16.h")
    cuda_fp16_hpp_abspath = get_include_file_abspath("cuda_fp16.hpp")
    if cuda_fp16_h_abspath and cuda_fp16_hpp_abspath:
        path = os.path.dirname(cuda_fp16_h_abspath)
        if path != os.path.dirname(cuda_fp16_hpp_abspath):
            KeOps_Error("cuda_fp16.h and cuda_fp16.hpp are not in the same folder !")
        return path
    else:
        KeOps_Error("cuda_fp16.h and cuda_fp16.hpp were not found")


def get_cuda_version(out_type="single_value"):
    cuda = ctypes.CDLL(find_library("cudart"))
    cuda_version = ctypes.c_int()
    cuda.cudaRuntimeGetVersion(ctypes.byref(cuda_version))
    cuda_version = int(cuda_version.value)
    if out_type == "single_value":
        return cuda_version
    cuda_version_major = cuda_version // 1000
    cuda_version_minor = (cuda_version - (1000 * cuda_version_major)) // 10
    if out_type == "major,minor":
        return cuda_version_major, cuda_version_minor
    elif out_type == "string":
        return f"{cuda_version_major}.{cuda_version_minor}"


def get_gpu_props():
    """
    Return number of GPU by reading libcuda.
    Here we assume the system has cuda support (more precisely that libcuda can be loaded)
    Adapted from https://gist.github.com/f0k/0d6431e3faa60bffc788f8b4daa029b1
    credit: Jan Schlüter
    """
    cuda = ctypes.CDLL(find_library("cuda"))

    nGpus = ctypes.c_int()
    error_str = ctypes.c_char_p()

    result = cuda.cuInit(0)
    if result != CUDA_SUCCESS:
        # cuda.cuGetErrorString(result, ctypes.byref(error_str))
        # KeOps_Warning("cuInit failed with error code %d: %s" % (result, error_str.value.decode()))
        KeOps_Warning(
            "cuda was detected, but driver API could not be initialized. Switching to cpu only."
        )
        return 0, ""

    result = cuda.cuDeviceGetCount(ctypes.byref(nGpus))
    if result != CUDA_SUCCESS:
        # cuda.cuGetErrorString(result, ctypes.byref(error_str))
        # KeOps_Warning("cuDeviceGetCount failed with error code %d: %s" % (result, error_str.value.decode()))
        KeOps_Warning(
            "cuda was detected, driver API has been initialized, but no working GPU has been found. Switching to cpu only."
        )
        return 0, ""

    nGpus = nGpus.value

    def safe_call(d, result):
        test = result == CUDA_SUCCESS
        if not test:
            KeOps_Warning(
                f"""
                    cuda was detected, driver API has been initialized, 
                    but there was an error for detecting properties of GPU device nr {d}. 
                    Switching to cpu only.
                """
            )
        return test

    test = True
    MaxThreadsPerBlock = [0] * (nGpus)
    SharedMemPerBlock = [0] * (nGpus)
    for d in range(nGpus):

        # getting handle to cuda device
        device = ctypes.c_int()
        result &= safe_call(d, cuda.cuDeviceGet(ctypes.byref(device), ctypes.c_int(d)))

        # getting MaxThreadsPerBlock info for device
        output = ctypes.c_int()
        result &= safe_call(
            d,
            cuda.cuDeviceGetAttribute(
                ctypes.byref(output),
                ctypes.c_int(CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_BLOCK),
                device,
            ),
        )
        MaxThreadsPerBlock[d] = output.value

        # getting SharedMemPerBlock info for device
        result &= safe_call(
            d,
            cuda.cuDeviceGetAttribute(
                ctypes.byref(output),
                ctypes.c_int(CU_DEVICE_ATTRIBUTE_MAX_SHARED_MEMORY_PER_BLOCK),
                device,
            ),
        )
        SharedMemPerBlock[d] = output.value

    # Building compile flags in the form "-D..." options for further compilations
    # (N.B. the purpose is to avoid the device query at runtime because it would slow down computations)
    string_flags = f"-DMAXIDGPU={nGpus-1} "
    for d in range(nGpus):
        string_flags += f"-DMAXTHREADSPERBLOCK{d}={MaxThreadsPerBlock[d]} "
        string_flags += f"-DSHAREDMEMPERBLOCK{d}={SharedMemPerBlock[d]} "

    if test:
        return nGpus, string_flags
    else:
        return 0, 0, ""



from keopscore.formulas.Operation import Operation
from keopscore.utils.code_gen_utils import c_zero_float
from keopscore.utils.misc_utils import KeOps_Error

# //////////////////////////////////////////////////////////////
# ////       ONE-HOT REPRESENTATION : OneHot<F,DIM>         ////
# //////////////////////////////////////////////////////////////


class OneHot(Operation):

    string_id = "OneHot"

    def __init__(self, f, dim):
        if f.dim != 1:
            KeOps_Error("One-hot representation is only supported for scalar formulas.")
        if dim < 1:
            KeOps_Error("A one-hot vector should have length >= 1.")
        super().__init__(f)
        self.dim = dim
        self.params = (dim,)

    def Op(self, out, table, arg0):
        if out.dtype == "half2" and arg0.dtype == "half2":
            return f"""
                        #pragma unroll
                        for (int k = 0; k < {self.dim}; k++)
                            {out.id}[k] = __heq2(h2rint(*{arg0.id}),__float2half2_rn(k));
                    """
        else:
            string = out.assign(c_zero_float)
            string += f"""
                           {out.id}[(int)(*{arg0.id}+.5)] = 1.0;
                       """
            return string

    def DiffT(self, v, gradin):
        from keopscore.formulas.variables.Zero import Zero

        return Zero(v.dim)



from keopscore.formulas.maths.SqNormDiag import SqNormDiag
from keopscore.formulas.maths.SqNormIso import SqNormIso
from keopscore.formulas.maths.SymSqNorm import SymSqNorm


class WeightedSqNorm:
    """
    WeightedSqNorm(A,X) : redirects to SqNormIso, SqNormDiag or SymSqNorm depending on dimension of A.
    """

    string_id = "WeightedSqNorm"

    def __new__(cls, A, X):
        if A.dim == 1:
            return SqNormIso(A, X)
        elif A.dim == X.dim:
            return SqNormDiag(A, X)
        else:
            return SymSqNorm(A, X)

    enable_test = True
    nargs = 2  # number of arguments
    test_argdims = [5, 5]  # dimensions of arguments for testing



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.formulas.maths.Sign import Sign
from keopscore.utils.math_functions import keops_abs


class Abs(VectorizedScalarOp):
    """the absolute value vectorized operation"""

    string_id = "Abs"

    ScalarOpFun = keops_abs

    @staticmethod
    def Derivative(f):
        return Sign(f)



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.formulas.maths.IntInv import IntInv


##########################
######    Rsqrt      #####
##########################


class Rsqrt(VectorizedScalarOp):
    """the inverse square root vectorized operation"""

    string_id = "Rsqrt"

    def ScalarOp(self, out, arg):
        from keopscore.utils.math_functions import keops_rsqrt

        # returns the atomic piece of c++ code to evaluate the function on arg and return
        # the result in out
        return out.assign(keops_rsqrt(arg))

    @staticmethod
    def Derivative(f):
        return IntInv(-2) * Rsqrt(f) ** 3

    # parameters for testing the operation (optional)
    test_ranges = [(0.5, 2)]  # range of argument



from keopscore.formulas.Chunkable_Op import Chunkable_Op
from keopscore.formulas.maths.Sum import Sum
from keopscore.utils.code_gen_utils import (
    c_zero_float,
    VectApply,
)
from keopscore.utils.math_functions import keops_fma
from keopscore.utils.misc_utils import KeOps_Error

##########################
#####    Scalprod     ####
##########################


class Scalprod_Impl(Chunkable_Op):
    string_id = "Scalprod"
    print_spec = "|", "mid", 3

    dim = 1

    def __init__(self, fa, fb):
        # Output dimension = 1, provided that FA::DIM = FB::DIM
        self.dimin = fa.dim
        if self.dimin != fb.dim:
            KeOps_Error("Dimensions must be the same for Scalprod")
        super().__init__(fa, fb)

    def Op(self, out, table, arga, argb):
        return out.assign(c_zero_float) + VectApply(self.ScalarOp, out, arga, argb)

    def ScalarOp(self, out, arga, argb):
        return out.assign(keops_fma(arga, argb, out))

    def DiffT(self, v, gradin):
        fa, fb = self.children
        return gradin * (fa.DiffT(v, fb) + fb.DiffT(v, fa))

    def initacc_chunk(self, acc):
        return f"*{acc.id} = 0.0f;\n"

    def acc_chunk(self, acc, out):
        return f"*{acc.id} += *{out.id};\n"


def Scalprod(arg0, arg1):
    if arg0.dim == 1:
        return arg0 * Sum(arg1)
    elif arg1.dim == 1:
        return Sum(arg0) * arg1
    else:
        return Scalprod_Impl(arg0, arg1)



from keopscore.binders.nvrtc.Gpu_link_compile import Gpu_link_compile
from keopscore.mapreduce.gpu.GpuAssignZero import GpuAssignZero
from keopscore.mapreduce.MapReduce import MapReduce
from keopscore.utils.code_gen_utils import (
    c_variable,
    c_array,
)


class GpuReduc1D(MapReduce, Gpu_link_compile):
    # class for generating the final C++ code, Gpu version

    AssignZero = GpuAssignZero

    def __init__(self, *args):
        MapReduce.__init__(self, *args)
        Gpu_link_compile.__init__(self)
        self.dimy = self.varloader.dimy

    def get_code(self):
        super().get_code()

        red_formula = self.red_formula
        dtype = self.dtype
        varloader = self.varloader

        i = self.i
        j = self.j
        fout = self.fout
        outi = self.outi
        acc = self.acc
        arg = self.arg
        args = self.args
        sum_scheme = self.sum_scheme

        param_loc = self.param_loc
        xi = self.xi
        yjloc = c_array(dtype, varloader.dimy, f"(yj + threadIdx.x * {varloader.dimy})")
        yjrel = c_array(dtype, varloader.dimy, "yjrel")
        table = varloader.table(self.xi, yjrel, self.param_loc)
        jreltile = c_variable("int", "(jrel + tile * blockDim.x)")

        self.code = f"""
                          
                        {self.headers}
                        
                        extern "C" __global__ void GpuConv1DOnDevice(int nx, int ny, {dtype} *out, {dtype} **{arg.id}) {{
    
                          // get the index of the current thread
                          int i = blockIdx.x * blockDim.x + threadIdx.x;

                          // declare shared mem
                          extern __shared__ {dtype} yj[];

                          // load parameters variables from global memory to local thread memory
                          {param_loc.declare()}
                          {varloader.load_vars("p", param_loc, args)}

                          {fout.declare()}
                          {xi.declare()}
                          {acc.declare()}
                          {sum_scheme.declare_temporary_accumulator()}

                          if (i < nx) {{
                            {red_formula.InitializeReduction(acc)} // acc = 0
                            {sum_scheme.initialize_temporary_accumulator_first_init()}
                            {varloader.load_vars('i', xi, args, row_index=i)} // load xi variables from global memory to local thread memory
                          }}

                          for (int jstart = 0, tile = 0; jstart < ny; jstart += blockDim.x, tile++) {{

                            // get the current column
                            int j = tile * blockDim.x + threadIdx.x;

                            if (j < ny) {{ // we load yj from device global memory only if j<ny
                              {varloader.load_vars("j", yjloc, args, row_index=j)} 
                            }}
                            __syncthreads();

                            if (i < nx) {{ // we compute x1i only if needed
                              {dtype} * yjrel = yj;
                              {sum_scheme.initialize_temporary_accumulator_block_init()}
                              for (int jrel = 0; (jrel < blockDim.x) && (jrel < ny - jstart); jrel++, yjrel += {varloader.dimy}) {{
                                {red_formula.formula(fout, table)} // Call the function, which outputs results in fout
                                {sum_scheme.accumulate_result(acc, fout, jreltile)}
                              }}
                              {sum_scheme.final_operation(acc)}
                            }}
                            __syncthreads();
                          }}
                          if (i < nx) {{
                            {red_formula.FinalizeOutput(acc, outi, i)} 
                          }}

                        }}
                    """



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.formulas.maths.Cos import Cos
from keopscore.formulas.maths.Sin import Sin
from keopscore.utils.math_functions import keops_sinxdivx


class SinXDivX(VectorizedScalarOp):
    """
    the sin(x)/x vectorized operation
    """

    string_id = "SinXDivX"

    ScalarOpFun = keops_sinxdivx

    @staticmethod
    def Derivative(f):
        return Cos(f) / f - Sin(f) / f**2



from keopscore.formulas.Operation import Broadcast
from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.formulas.maths.Scalprod import Scalprod
from keopscore.formulas.maths.Square import Square
from keopscore.formulas.variables.Zero import Zero
from keopscore.utils.math_functions import keops_mul
from keopscore.formulas.variables.IntCst import IntCst_Impl
from keopscore.formulas.maths.SumT import SumT, SumT_Impl
from keopscore.utils.misc_utils import KeOps_Error

##########################
######    Mult       #####
##########################


class Mult_Impl(VectorizedScalarOp):
    """the binary multiply operation"""

    string_id = "Mult"
    print_spec = "*", "mid", 3

    ScalarOpFun = keops_mul

    #  \diff_V (A*B) = (\diff_V A) * B + A * (\diff_V B)
    def DiffT(self, v, gradin):
        fa, fb = self.children
        if fa.dim == 1 and fb.dim > 1:
            return fa.DiffT(v, Scalprod(gradin, fb)) + fb.DiffT(v, fa * gradin)
        elif fb.dim == 1 and fa.dim > 1:
            return fa.DiffT(v, fb * gradin) + fb.DiffT(v, Scalprod(gradin, fa))
        else:
            return fa.DiffT(v, fb * gradin) + fb.DiffT(v, fa * gradin)

    # parameters for testing the operation (optional)
    nargs = 2  # number of arguments
    torch_op = "torch.mul"  # equivalent PyTorch operation


# N.B. The following separate function should theoretically be implemented
# as a __new__ method of the previous class, but this can generate infinite recursion problems
def Mult(arg0, arg1):
    if isinstance(arg0, Zero):
        return Broadcast(arg0, arg1.dim)
    elif isinstance(arg1, Zero):
        return Broadcast(arg1, arg0.dim)
    elif isinstance(arg0, IntCst_Impl):
        if arg0.val == 1:
            # 1*f -> f
            return arg1
        if arg0.val == -1:
            # -1*f -> -f
            return -arg1
        elif isinstance(arg1, IntCst_Impl):
            # m*n -> mn
            return IntCst_Impl(arg0.val * arg1.val)
    if isinstance(arg1, IntCst_Impl):
        # f*n -> n*f (bringing integers to the left)
        return Mult(arg1, arg0)
    elif isinstance(arg1, Mult_Impl) and isinstance(arg1.children[0], IntCst_Impl):
        # f*(n*g) -> (n*f)*g
        return (arg1.children[0] * arg0) * arg1.children[1]
    elif arg0 == arg1:
        # f*f -> f^2
        return Square(arg0)
    elif isinstance(arg1, SumT_Impl):
        # f*SumT(g)
        if isinstance(arg0, SumT_Impl):
            # SumT(f)*SumT(g) -> SumT(f*g)
            if arg0.dim != arg1.dim:  # should never happen...
                KeOps_Error("dimensions are not compatible for Mult operation")
            return SumT(arg0.children[0] * arg1.children[0], arg0.dim)
        elif arg0.dim == 1:
            # f*SumT(g) -> SumT(f*g)
            return SumT(arg0 * arg1.children[0], arg1.dim)
        elif arg1.dim == arg0.dim:
            # f*SumT(g) -> f*g (broadcasted)
            return arg0 * arg1.children[0]
        else:
            KeOps_Error("dimensions are not compatible for Mult operation")
    elif isinstance(arg0, SumT_Impl):
        # SumT(f)*g -> g*SumT(f)
        return arg1 * arg0
    else:
        return Mult_Impl(arg0, arg1)



from keopscore import cuda_block_size
from keopscore.config.chunks import dimfinalchunk
from keopscore.binders.nvrtc.Gpu_link_compile import Gpu_link_compile
from keopscore.formulas.reductions.Sum_Reduction import Sum_Reduction
from keopscore.formulas.reductions.sum_schemes import *
from keopscore.mapreduce.gpu.GpuAssignZero import GpuAssignZero
from keopscore.mapreduce.MapReduce import MapReduce
from keopscore.utils.code_gen_utils import (
    load_vars,
    load_vars_chunks,
    load_vars_chunks_offsets,
    sizeof,
    pointer,
    Var_loader,
    use_pragma_unroll,
)
from keopscore.utils.misc_utils import KeOps_Error


def do_finalchunk_sub_ranges(
    dtype,
    fun_global,
    varfinal,
    dimfinalchunk_curr,
    acc,
    i,
    j,
    jstart,
    start_y,
    chunk,
    end_x,
    end_y,
    nbatchdims,
    indices_j,
    arg,
    fout,
    yj,
    out,
):

    dimout = varfinal.dim
    yjloc = c_variable(pointer(dtype), f"({yj.id} + threadIdx.x * {dimfinalchunk})")
    indsj_global = Var_loader(fun_global).indsj
    load_chunks_routine_j = load_vars_chunks(
        [varfinal.ind],
        dimfinalchunk,
        dimfinalchunk_curr,
        varfinal.dim,
        yjloc,
        arg,
        chunk,
        row_index=j,
    )
    load_chunks_routine_j_ranges = load_vars_chunks_offsets(
        [varfinal.ind],
        indsj_global,
        dimfinalchunk,
        dimfinalchunk_curr,
        varfinal.dim,
        yjloc,
        arg,
        chunk,
        indices_j,
        row_index=j - start_y,
    )
    return f"""
                {acc.assign(c_zero_float)}
                {dtype} *yjrel = yj;
                if ({j.id} < {end_y.id}) {{ // we load yj from device global memory only if j<end_y
                    if ({nbatchdims.id}==0) {{
                        {load_chunks_routine_j}
                    }} else {{
                        {load_chunks_routine_j_ranges}
                    }}
                }}
                __syncthreads();
                for (int jrel = 0; (jrel < blockDim.x) && (jrel < {end_y.id} - {jstart.id}); jrel++, yjrel += {dimfinalchunk}) {{          
                    if ({i.id} < {end_x.id}) {{ // we compute only if needed
                        {use_pragma_unroll()}
                        for (int k=0; k<{dimfinalchunk_curr}; k++) {{
                            {acc.id}[k] += yjrel[k] * fout[jrel];
                        }}
                    }}
                    __syncthreads();
                }}
                if ({i.id} < {end_x.id}) {{
                    {use_pragma_unroll()}
                    for (int k=0; k<{dimfinalchunk_curr}; k++)
                        {out.id}[i*{dimout}+{chunk.id}*{dimfinalchunk}+k] += {acc.id}[k];
                }}
                __syncthreads();
            """


class GpuReduc1D_ranges_finalchunks(MapReduce, Gpu_link_compile):
    # class for generating the final C++ code, Gpu version

    AssignZero = GpuAssignZero

    def __init__(self, *args):
        MapReduce.__init__(self, *args)
        Gpu_link_compile.__init__(self)

    def get_code(self):

        super().get_code()

        dtype = self.dtype
        dtypeacc = self.dtypeacc
        i = self.i
        j = self.j
        nx = c_variable("int", "nx")
        ny = c_variable("int", "ny")
        jstart = c_variable("int", "jstart")
        chunk = c_variable("int", "chunk")
        arg = self.arg
        args = self.args
        yj = c_variable(pointer(dtype), "yj")
        out = c_variable(pointer(dtype), "out")

        fun_internal = Sum_Reduction(
            self.red_formula.formula.children[0], self.red_formula.tagI
        )
        formula = fun_internal.formula
        varfinal = self.red_formula.formula.children[1]
        nchunks = 1 + (varfinal.dim - 1) // dimfinalchunk
        dimlastfinalchunk = varfinal.dim - (nchunks - 1) * dimfinalchunk
        varloader = Var_loader(fun_internal)
        dimsx = varloader.dimsx
        dimsy = varloader.dimsy
        dimsp = varloader.dimsp
        indsi = varloader.indsi
        indsj = varloader.indsj
        indsp = varloader.indsp
        dimx = sum(dimsx)
        dimy = sum(dimsy)
        dimp = sum(dimsp)
        dimout = varfinal.dim
        dimfout = fun_internal.formula.dim
        if dimfout != 1:
            KeOps_Error("dimfout should be 1")
        sum_scheme = self.sum_scheme

        self.dimy = max(dimfinalchunk, dimy)
        blocksize_chunks = min(
            cuda_block_size, 1024, 49152 // max(1, self.dimy * sizeof(self.dtype))
        )

        if not isinstance(sum_scheme, block_sum):
            KeOps_Error("only block_sum available")
        param_loc = c_array(dtype, dimp, "param_loc")
        fout = c_array(dtype, dimfout * blocksize_chunks, "fout")
        xi = c_array(dtype, dimx, "xi")
        acc = c_array(dtypeacc, dimfinalchunk, "acc")
        yjloc = c_array(dtype, dimy, f"(yj + threadIdx.x * {dimy})")
        foutjrel = c_array(dtype, dimfout, f"({fout.id}+jrel*{dimfout})")
        yjrel = c_array(dtype, dimy, "yjrel")
        table = varloader.table(xi, yjrel, param_loc)

        lastchunk = c_variable("int", f"{nchunks-1}")

        startx = c_variable("int", "start_x")
        starty = c_variable("int", "start_y")

        end_x = c_variable("int", "end_x")
        end_y = c_variable("int", "end_y")

        nbatchdims = c_variable("int", "nbatchdims")

        fun_global = self.red_formula
        varloader_global = Var_loader(fun_global)
        indsi_global = varloader_global.indsi
        indsj_global = varloader_global.indsj
        nvarsi_global, nvarsj_global, nvarsp_global = (
            len(varloader_global.Varsi),
            len(varloader_global.Varsj),
            len(varloader_global.Varsp),
        )
        nvars_global = nvarsi_global + nvarsj_global + nvarsp_global
        offsets = c_array("int", nvars_global, "offsets")

        indices_i = c_array("int", nvarsi_global, "indices_i")
        indices_j = c_array("int", nvarsj_global, "indices_j")
        indices_p = c_array("int", nvarsp_global, "indices_p")

        declare_assign_indices_i = (
            "int *indices_i = offsets;" if nvarsi_global > 0 else ""
        )
        declare_assign_indices_j = (
            f"int *indices_j = offsets + {nvarsi_global};" if nvarsj_global > 0 else ""
        )
        declare_assign_indices_p = (
            f"int *indices_p = offsets + {nvarsi_global} + {nvarsj_global};"
            if nvarsp_global > 0
            else ""
        )

        chunk_sub_routine = do_finalchunk_sub_ranges(
            dtype,
            fun_global,
            varfinal,
            dimfinalchunk,
            acc,
            i,
            j,
            jstart,
            starty,
            chunk,
            end_x,
            end_y,
            nbatchdims,
            indices_j,
            arg,
            fout,
            yj,
            out,
        )

        chunk_sub_routine_last = do_finalchunk_sub_ranges(
            dtype,
            fun_global,
            varfinal,
            dimlastfinalchunk,
            acc,
            i,
            j,
            jstart,
            starty,
            lastchunk,
            end_x,
            end_y,
            nbatchdims,
            indices_j,
            arg,
            fout,
            yj,
            out,
        )

        threadIdx_x = c_variable("int", "threadIdx.x")

        self.code = f"""
                          
                        {self.headers}
                        
                        extern "C" __global__ void GpuConv1DOnDevice_ranges(int nx, int ny, int nbatchdims,
                                                    int *offsets_d, int *lookup_d, int *slices_x,
                                                    int *ranges_y, {dtype} *out, {dtype} **{arg.id}) {{
                                                        
                          {offsets.declare()}
                          {declare_assign_indices_i}
                          {declare_assign_indices_j}
                          {declare_assign_indices_p}
                          
                          if (nbatchdims > 0) {{
                              for (int k = 0; k < {nvars_global}; k++) {{
                                  offsets[k] = offsets_d[ {nvars_global} * blockIdx.x + k ];
                              }}
                          }}
                       
                          // Retrieve our position along the laaaaarge [1,~nx] axis: -----------------
                          int range_id= (lookup_d)[3*blockIdx.x] ;
                          int start_x = (lookup_d)[3*blockIdx.x+1] ;
                          int end_x   = (lookup_d)[3*blockIdx.x+2] ;
                          
                          // The "slices_x" vector encodes a set of cutting points in
                          // the "ranges_y" array of ranges.
                          // As discussed in the Genred docstring, the first "0" is implicit:
                          int start_slice = range_id < 1 ? 0 : slices_x[range_id-1];
                          int end_slice   = slices_x[range_id];
                          
                          // get the index of the current thread
                          int i = start_x + threadIdx.x;
                          
                          // declare shared mem
                          extern __shared__ {dtype} yj[];
                          
                          // load parameter(s)
                          {param_loc.declare()}
                          {load_vars(dimsp, indsp, param_loc, args)}
                          
                          {fout.declare()}
    
                          // get the value of variable (index with i)
                          {xi.declare()}
                          if (i < end_x) {{
                              
                              if (nbatchdims == 0) {{
                                  {varloader.load_vars("i", xi, args, row_index=i)} // load xi variables from global memory to local thread memory
                              }} else {{
                                  {varloader.load_vars("i", xi, args, row_index=threadIdx_x, offsets=indices_i, indsref=indsi_global)} // Possibly, with offsets as we support broadcasting over batch dimensions
                              }}
                              
                              {use_pragma_unroll()}
                              for (int k=0; k<{dimout}; k++) {{
                                  out[i*{dimout}+k] = 0.0f;
                              }}
                          }}
                          
                          {acc.declare()}
                          
                          int start_y = ranges_y[2*start_slice], end_y = 0;
                          for(int index = start_slice ; index < end_slice ; index++ ) {{
                              if( (index+1 >= end_slice) || (ranges_y[2*index+2] != ranges_y[2*index+1]) ) {{
                                  end_y = ranges_y[2*index+1];
                                  for (int jstart = start_y, tile = 0; jstart < end_y; jstart += blockDim.x, tile++) {{
                                      
                                      // get the current column
                                      int j = jstart + threadIdx.x;
                                      
                                      if (j < end_y) {{ // we load yj from device global memory only if j<end_y
                                          if (nbatchdims == 0) {{
                                              {varloader.load_vars("j", yjloc, args, row_index=j)} // load yj variables from global memory to shared memory
                                          }} else {{
                                              {varloader.load_vars("j", yjloc, args, row_index=j-starty, offsets=indices_j, indsref=indsj_global)} // Possibly, with offsets as we support broadcasting over batch dimensions
                                          }}
                                      }}
                                      __syncthreads();
                                      
                                      if (i < end_x) {{ // we compute x1i only if needed
                                          {dtype} * yjrel = yj; // Loop on the columns of the current block.
                                          for (int jrel = 0; (jrel < {blocksize_chunks}) && (jrel < end_y - jstart); jrel++, yjrel += {dimy}) {{
                                              {formula(foutjrel, table)} // Call the function, which outputs results in fout
                                          }}
                                      }}
                                      __syncthreads();
                                      
                                      for (int chunk=0; chunk<{nchunks-1}; chunk++) {{
                                          {chunk_sub_routine}
                                      }}
                                      {chunk_sub_routine_last}
                                  }}
                                  if(index+1 < end_slice) 
                                      start_y = ranges_y[2*index+2];
                              }}
                          }}
                       }}
                    """



from keopscore.formulas.reductions.KMin_ArgKMin_Reduction import KMin_ArgKMin_Reduction
from keopscore.utils.code_gen_utils import (
    infinity,
    cast_to,
    c_zero_float,
    c_for_loop,
    c_variable,
    new_c_varname,
    c_if,
)


class KMin_Reduction(KMin_ArgKMin_Reduction):
    """Implements the k-min reduction operation : for each i or each j, find the
    k minimal values of Fij operation is vectorized: if Fij is vector-valued, arg-k-min
    is computed for each dimension."""

    string_id = "KMin_Reduction"

    def __init__(self, formula, K, tagIJ):
        super().__init__(formula, K, tagIJ)
        self.dim = K * formula.dim

    def FinalizeOutput(self, acc, out, i):
        fdim, K = self.formula.dim, self.K
        outer_loop, k = c_for_loop(0, fdim, 1)
        inner_loop, l = c_for_loop(k, k + (2 * fdim * K), 2 * fdim)
        p = c_variable("int", new_c_varname("p"))
        return outer_loop(
            p.declare_assign(k) + inner_loop(out[p].assign(acc[l]) + p.add_assign(fdim))
        )



from keopscore.utils.code_gen_utils import cast_to, c_variable
from keopscore.formulas.Operation import Operation
from keopscore.formulas.variables.Zero import Zero


class IntCst_Impl(Operation):
    # constant integer "operation"
    string_id = "IntCst"
    print_spec = "", "pre", 0

    def __init__(self, val):
        super().__init__()
        self.val = val
        self.dim = 1
        self.params = (val,)

    # custom __eq__ method
    def __eq__(self, other):
        return type(self) == type(other) and self.val == other.val

    def Op(self, out, table):
        float_val = c_variable("float", f"(float){self.val}")
        return f"*{out.id} = {cast_to(out.dtype, float_val)};\n"

    def DiffT(self, v, gradin):
        return Zero(v.dim)


# N.B. The following separate function should theoretically be implemented
# as a __new__ method of the previous class, but this can generate infinite recursion problems
def IntCst(arg):
    if arg == 0:
        return Zero(1)
    else:
        return IntCst_Impl(arg)



from keopscore.formulas.maths.ElemT import ElemT
from keopscore.formulas.maths.Concat import Concat
from keopscore.formulas.variables.IntCst import IntCst
from keopscore.formulas.variables.Var import Var
from keopscore.utils.code_gen_utils import GetInds


# //////////////////////////////////////////////////////////////////////////////////////////////
# ////      Standard basis of R^DIM : < (1,0,0,...) , (0,1,0,...) , ... , (0,...,0,1) >     ////
# //////////////////////////////////////////////////////////////////////////////////////////////


def StandardBasis(dim):
    return tuple(ElemT(IntCst(1), dim, i) for i in range(dim))


# /////////////////////////////////////////////////////////////////////////
# ////      Matrix of gradient operator (=transpose of jacobian)       ////
# /////////////////////////////////////////////////////////////////////////


class GradMatrix:
    def __new__(cls, f, v):
        f.Vars(cat=3)
        IndsTempVars = GetInds(f.Vars(cat=3))
        newind = 1 if len(IndsTempVars) == 0 else 1 + max(IndsTempVars)
        gradin = Var(newind, f.dim, 3)
        packGrads = tuple(
            f.DiffT(v, gradin).replace(gradin, e) for e in StandardBasis(f.dim)
        )
        res = packGrads[0]
        for elem in packGrads[1:]:
            res = Concat(res, elem)
        return res

    enable_test = False



#######################################################################
# .  Warnings, Errors, etc.
#######################################################################

import keopscore


def KeOps_Message(message, use_tag=True, **kwargs):
    if keopscore.verbose:
        tag = "[KeOps] " if use_tag else ""
        message = tag + message
        print(message, **kwargs)


def KeOps_Warning(message):
    if keopscore.verbose:
        message = "[KeOps] Warning : " + message
        print(message)


def KeOps_Error(message, show_line_number=True):
    message = "[KeOps] Error : " + message
    if show_line_number:
        from inspect import currentframe, getframeinfo

        frameinfo = getframeinfo(currentframe().f_back)
        message += f" (error at line {frameinfo.lineno} in file {frameinfo.filename})"
    raise ValueError(message)


def KeOps_OS_Run(command):
    import sys

    python_version = sys.version_info
    if python_version >= (3, 7):
        import subprocess

        out = subprocess.run(command, shell=True, capture_output=True)
        if out.stderr != b"":
            print(out.stderr.decode("utf-8"))
            KeOps_Error("Error compiling formula.")
    elif python_version >= (3, 5):
        import subprocess

        subprocess.run(
            command,
            shell=True,
        )
    else:
        import os

        os.system(command)


def find_library_abspath(lib):
    """
    wrapper around ctypes find_library that returns the full path
    of the library.
    Warning : it also opens the shared library !
    Adapted from
    https://stackoverflow.com/questions/35682600/get-absolute-path-of-shared-library-in-python/35683698
    """
    from ctypes import c_int, c_void_p, c_char_p, CDLL, byref, cast, POINTER, Structure
    from ctypes.util import find_library

    # linkmap structure, we only need the second entry
    class LINKMAP(Structure):
        _fields_ = [("l_addr", c_void_p), ("l_name", c_char_p)]

    res = find_library(lib)
    if res is None:
        return ""

    lib = CDLL(res)
    libdl = CDLL(find_library("dl"))

    dlinfo = libdl.dlinfo
    dlinfo.argtypes = c_void_p, c_int, c_void_p
    dlinfo.restype = c_int

    # gets typecasted later, I dont know how to create a ctypes struct pointer instance
    lmptr = c_void_p()

    # 2 equals RTLD_DI_LINKMAP, pass pointer by reference
    dlinfo(lib._handle, 2, byref(lmptr))

    # typecast to a linkmap pointer and retrieve the name.
    abspath = cast(lmptr, POINTER(LINKMAP)).contents.l_name

    return abspath.decode("utf-8")



from keopscore.formulas.Operation import Operation
from keopscore.utils.code_gen_utils import c_for_loop
from keopscore.utils.misc_utils import KeOps_Error

# /////////////////////////////////////////////////////////////////////////
# ////      adjoint of ComplexSum                           ////
# /////////////////////////////////////////////////////////////////////////


class ComplexSumT(Operation):
    string_id = "ComplexSumT"

    def __init__(self, f, dim):
        if f.dim != 2:
            KeOps_Error("Dimension of F must be 2")
        self.dim = dim
        super().__init__(f)

    def Op(self, out, table, inF):
        forloop, i = c_for_loop(0, self.dim, 2, pragma_unroll=True)
        body = out[i].assign(inF[0])
        body += out[i + 1].assign(inF[1])
        return forloop(body)

    def DiffT(self, v, gradin):
        from keopscore.formulas.complex.ComplexSum import ComplexSum

        f = self.children[0]
        return f.DiffT(v, ComplexSum(gradin))



from keopscore.formulas.maths.Sum import Sum
from keopscore.formulas.maths.TensorProd import TensorProd


class SymSqNorm:
    """
    Fully anisotropic norm, if S.dim == A.dim * A.dim
    SymSqNorm(A,X) = sum_{ij} a_ij * x_i*x_j
    """

    def __new__(cls, A, X):
        return Sum(A * TensorProd(X, X))

    enable_test = False



from keopscore.formulas.maths.Atan2 import Atan2
from keopscore.formulas.complex.ComplexReal import ComplexReal
from keopscore.formulas.complex.ComplexImag import ComplexImag

# /////////////////////////////////////////////////////////////////////////
# ////      ComplexAngle                           ////
# /////////////////////////////////////////////////////////////////////////


def ComplexAngle(f):
    return Atan2(ComplexImag(f), ComplexReal(f))



from keopscore.formulas import Var
from keopscore.utils.code_gen_utils import GetInds
from keopscore.utils.misc_utils import KeOps_Error

# /////////////////////////////////////////////////////////////
# ///      GRADIENT OPERATOR  : Grad< F, V, Gradin >       ////
# /////////////////////////////////////////////////////////////

# Defines [\partial_V F].gradin function
# Symbolic differentiation is a straightforward recursive operation,
# provided that the operators have implemented their DiffT "compiler methods":


def Grad(formula, v, gradin=None):
    if gradin is None:
        if v.cat == 2:
            KeOps_Error("not implemented")
        inds = GetInds(formula.Vars_)
        ind = 1 + max(inds) if len(inds) > 0 else 0
        dim = formula.dim
        cat = 1 - v.cat
        gradin = Var(ind, dim, cat)
    return formula.DiffT(v, gradin)



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.formulas.maths.Step import Step
from keopscore.utils.math_functions import keops_relu


class ReLU(VectorizedScalarOp):
    """the ReLU vectorized operation"""

    string_id = "ReLU"

    ScalarOpFun = keops_relu

    @staticmethod
    def Derivative(f):
        return Step(f)



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.utils.math_functions import keops_cos


class Cos(VectorizedScalarOp):
    """the cosine vectorized operation"""

    string_id = "Cos"

    ScalarOpFun = keops_cos

    @staticmethod
    def Derivative(f):
        from .Sin import Sin

        return -Sin(f)



from keopscore.formulas.Operation import Operation
from keopscore.formulas.variables.Zero import Zero
from keopscore.utils.code_gen_utils import (
    c_zero_float,
    c_for_loop,
    c_if,
    value,
    c_variable,
)
from keopscore.utils.misc_utils import KeOps_Error

############################
######    ArgMin       #####
############################


class ArgMin(Operation):
    string_id = "ArgMin"

    def __init__(self, f):
        super().__init__(f)
        if f.dim < 1:
            KeOps_Error("ArgMin operation is only possible when dimension is non zero.")
        self.dim = 1

    def Op(self, out, table, arg):
        tmp = c_variable(out.dtype)
        loop, k = c_for_loop(1, arg.dim, 1, pragma_unroll=True)
        string = value(out).assign(c_zero_float) + tmp.declare_assign(arg[0])
        if out.dtype == "half2":
            loop_string = f"""
                // we have to work element-wise...
                __half2 cond = __hgt2({tmp.id},{arg[k].id});                          // cond = (tmp > outF[k]) (element-wise)
                __half2 negcond = __float2half2_rn(1.0f)-cond;                        // negcond = 1-cond
                *{out.id} = cond * __float2half2_rn({k.id}) + negcond * *{out.id};    // out  = cond * k + (1-cond) * out 
                {tmp.id} = cond * {arg[k].id} + negcond * {tmp.id};                   // tmp  = cond * outF[k] + (1-cond) * tmp
                            """
            string += loop(loop_string)
        else:
            string += loop(
                c_if(arg[k] < tmp, tmp.assign(arg[k]) + value(out).assign(k))
            )
        return string

    def DiffT(self, v, gradin):
        return Zero(v.dim)

    # parameters for testing the operation (optional)
    enable_test = True  # enable testing for this operation
    nargs = 1  # number of arguments
    test_argdims = [5]  # dimensions of arguments for testing
    torch_op = "lambda x : torch.argmin(x, dim=-1, keepdim=True).type(x.dtype)"
    no_torch_grad = True



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.utils.math_functions import keops_sin


class Sin(VectorizedScalarOp):
    """the Sine vectorized operation"""

    string_id = "Sin"

    ScalarOpFun = keops_sin

    @staticmethod
    def Derivative(f):
        from .Cos import Cos

        return Cos(f)



from .ArgKMin_Reduction import ArgKMin_Reduction
from .ArgMax_Reduction import ArgMax_Reduction
from .ArgMin_Reduction import ArgMin_Reduction
from .KMin_ArgKMin_Reduction import KMin_ArgKMin_Reduction
from .KMin_Reduction import KMin_Reduction
from .Max_ArgMax_Reduction import Max_ArgMax_Reduction
from .Max_Reduction import Max_Reduction
from .Max_SumShiftExpWeight_Reduction import (
    Max_SumShiftExpWeight_Reduction,
    Max_SumShiftExp_Reduction,
)
from .Min_ArgMin_Reduction import Min_ArgMin_Reduction
from .Min_Reduction import Min_Reduction
from .Sum_Reduction import Sum_Reduction
from .Zero_Reduction import Zero_Reduction

from .sum_schemes import *



#  Always prefer setuptools over distutils
# To use a consistent encoding
from codecs import open
import os
from os import path

from setuptools import setup

here = path.abspath(path.dirname(__file__))

with open(os.path.join(here, "keopscore", "keops_version"), encoding="utf-8") as v:
    current_version = v.read().rstrip()

# Get the long description from the README file
with open(path.join(here, "keopscore", "readme.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="keopscore",
    version=current_version,
    description="keopscore is the KeOps meta programming engine. This python module should be used through a binder (e.g. pykeops or rkeops)",  # Required
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://www.kernel-operations.io/",
    project_urls={
        "Bug Reports": "https://github.com/getkeops/keops/issues",
        "Source": "https://github.com/getkeops/keops",
    },
    author="B. Charlier, J. Feydy, J. Glaunes",
    author_email="benjamin.charlier@umontpellier.fr, jean.feydy@gmail.com, alexis.glaunes@parisdescartes.fr",
    python_requires=">=3",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: C++",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords="kernels gpu autodiff",
    packages=[
        "keopscore",
        "keopscore.binders",
        "keopscore.binders.cpp",
        "keopscore.binders.nvrtc",
        "keopscore.config",
        "keopscore.formulas",
        "keopscore.formulas.autodiff",
        "keopscore.formulas.complex",
        "keopscore.formulas.maths",
        "keopscore.formulas.reductions",
        "keopscore.formulas.variables",
        "keopscore.include",
        "keopscore.mapreduce",
        "keopscore.mapreduce.cpu",
        "keopscore.mapreduce.gpu",
        "keopscore.tests",
        "keopscore.utils",
    ],
    package_data={
        "keopscore": [
            "readme.md",
            "licence.txt",
            "keops_version",
            "config/libiomp5.dylib",
            "binders/nvrtc/keops_nvrtc.cpp",
            "binders/nvrtc/nvrtc_jit.cpp",
            "include/CudaSizes.h",
            "include/ranges_utils.h",
            "include/Ranges.h",
            "include/Sizes.h",
            "include/utils_pe.h",
        ],
    },
    install_requires=[],
    extras_require={},
)



from keopscore.utils.code_gen_utils import c_zero_float
from keopscore.formulas.Operation import Operation


class Zero(Operation):
    """zero operation : encodes a vector of zeros"""

    string_id = "Zero"

    def __init__(self, dim):
        super().__init__()
        self.dim = dim
        self.params = (dim,)

    # custom __eq__ method
    def __eq__(self, other):
        return type(self) == type(other) and self.dim == other.dim

    def Op(self, out, table):
        return out.assign(c_zero_float)

    def DiffT(self, v, gradin):
        return Zero(v.dim)



from keopscore.formulas.Operation import Operation
from keopscore.formulas.maths.Extract import Extract
from keopscore.utils.code_gen_utils import VectCopy


############################
######    Concat       #####
############################


class Concat(Operation):
    string_id = "Concat"

    def __init__(self, arg0, arg1):
        super().__init__(arg0, arg1)
        self.dim = arg0.dim + arg1.dim

    def Op(self, out, table, arg0, arg1):
        out0, out1 = out.split(arg0.dim, arg1.dim)
        return VectCopy(out0, arg0) + VectCopy(out1, arg1)

    def DiffT(self, v, gradin):
        f = self.children[0]
        g = self.children[1]
        return f.DiffT(v, Extract(gradin, 0, f.dim)) + g.DiffT(
            v, Extract(gradin, f.dim, g.dim)
        )

    # parameters for testing the operation (optional)
    enable_test = True  # enable testing for this operation
    nargs = 2  # number of arguments
    test_argdims = [5, 3]  # dimensions of arguments for testing
    torch_op = None



import time
import torch
import numpy as np
from torch.autograd import grad
from pykeops.torch import Genred
from keopscore.formulas import *
import types


def TestOperation(op_str, tol=1e-4, dtype="float32", test_grad=True):

    # N.B. dtype can be 'float32', 'float64' or 'float16'

    print("")

    keops_op = eval(op_str)
    if isinstance(keops_op, types.FunctionType):
        op_class_str = f"{op_str}_Impl"
        exec(f"from {keops_op.__module__} import {op_class_str}")
        keops_op_class = eval(op_class_str)
    else:
        keops_op_class = keops_op

    if keops_op_class.enable_test:
        if hasattr(keops_op_class, "test_argdims"):
            dims = keops_op_class.test_argdims
            nargs = len(dims)
        else:
            if hasattr(keops_op_class, "nargs"):
                nargs = keops_op_class.nargs
            elif hasattr(keops_op_class, "test_ranges"):
                nargs = len(keops_op_class.test_ranges)
            elif hasattr(keops_op_class, "Derivative"):
                from inspect import signature

                nargs = len(signature(keops_op_class.Derivative).parameters)
                if hasattr(keops_op_class, "test_params"):
                    nargs -= len(keops_op_class.test_params)
            else:
                print("no test available for " + op_str)
                return None
            dims = [3] * nargs
    else:
        print("no test available for " + op_str)
        return None

    #####################################################################
    # Declare random inputs:

    M = 3000
    N = 5000

    # Choose the storage place for our data : CPU (host) or GPU (device) memory.
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    if dtype == "float32":
        torchtype = torch.float32
    elif dtype == "float64":
        torchtype = torch.float64
    elif dtype == "float16":
        torchtype = torch.float16
    else:
        raise ValueError("invalid dtype")

    argcats = np.random.choice(["i", "j"], nargs)

    if hasattr(keops_op_class, "test_ranges"):
        rng = keops_op_class.test_ranges
        rand = torch.rand
    else:
        rng = [(0, 1)] * nargs
        rand = torch.randn

    args = [None] * nargs
    for k in range(nargs):
        MorN = M if argcats[k] == "i" else N
        args[k] = rand(
            MorN, dims[k], dtype=torchtype, device=device, requires_grad=True
        )
        args[k] = args[k] * (rng[k][1] - rng[k][0]) + rng[k][0]

    ####################################################################
    # Define a custom formula
    # -----------------------

    if hasattr(keops_op_class, "test_params"):
        params = keops_op_class.test_params
    else:
        params = ()

    formula = (
        op_str
        + "("
        + ",".join(f"v{k}" for k in range(nargs))
        + ","
        + ",".join(str(p) for p in params)
        + ")"
    )

    variables = list(f"v{k} = V{argcats[k]}({dims[k]})" for k in range(nargs))

    print("Testing operation " + op_str)

    my_routine = Genred(formula, variables, reduction_op="Sum", axis=1)
    c = my_routine(*args)

    print("ok, no error")
    print("5 first values :", *c.flatten()[:5].tolist())

    ####################################################################
    # Compute the gradient
    # -----------------------

    if test_grad:
        e = torch.rand_like(c)

        print("Testing gradient of operation " + op_str)

        g = grad(c, args, e)

        print("ok, no error")
        for k in range(nargs):
            app_str = f"number {k}" if len(args) > 1 else ""
            print(
                f"5 first values for gradient {app_str}:", *g[k].flatten()[:5].tolist()
            )

    if not hasattr(keops_op_class, "torch_op"):
        torch_op_str = keops_op_class.string_id.lower()
        if not torch_op_str in dir(torch):
            return "no_torch"
        torch_op = "torch." + torch_op_str
    else:
        if keops_op_class.torch_op is None:
            return "no_torch"
        torch_op = keops_op_class.torch_op

    print("Comparing with PyTorch implementation ")

    torch_args = [None] * nargs
    for k in range(nargs):
        torch_args[k] = (
            args[k][:, None, :] if argcats[k] == "i" else args[k][None, :, :]
        )

    ####################################################################
    # The equivalent code with a "vanilla" pytorch implementation

    if isinstance(torch_op, str):
        torch_op = eval(torch_op)

    c_torch = torch_op(*torch_args, *params).sum(dim=1)
    err_op = torch.norm(c - c_torch).item() / torch.norm(c_torch).item()
    print("relative error for operation :", err_op)

    if test_grad:
        if (
            not hasattr(keops_op_class, "no_torch_grad")
            or not keops_op_class.no_torch_grad
        ):
            g_torch = grad(c_torch, args, e)
            err_gr = [None] * nargs
            for k in range(nargs):
                app_str = f"number {k}" if len(args) > 1 else ""
                err_gr[k] = (
                    torch.norm(g[k] - g_torch[k]) / torch.norm(g_torch[k])
                ).item()
                print(f"relative error for gradient {app_str}:", err_gr[k])
        else:
            print("no gradient for torch")
            return abs(err_op) > tol
        return abs(err_op) > tol, list(abs(err) > tol for err in err_gr)
    else:
        return abs(err_op) > tol



import os
import pickle
import keopscore

# global configuration parameter to be added for the lookup :
env_param = keopscore.config.config.cpp_flags


class Cache:
    def __init__(self, fun, use_cache_file=False, save_folder="."):
        self.fun = fun
        self.library = {}
        self.use_cache_file = use_cache_file
        if use_cache_file:
            self.cache_file = os.path.join(save_folder, fun.__name__ + "_cache.pkl")
            if os.path.isfile(self.cache_file) and os.path.getsize(self.cache_file) > 0:
                f = open(self.cache_file, "rb")
                self.library = pickle.load(f)
                f.close()
            import atexit

            atexit.register(self.save_cache)

    def __call__(self, *args):
        str_id = "".join(list(str(arg) for arg in args)) + str(env_param)
        if not str_id in self.library:
            self.library[str_id] = self.fun(*args)
        return self.library[str_id]

    def reset(self, new_save_folder=None):
        self.library = {}
        if new_save_folder:
            self.save_folder = new_save_folder

    def save_cache(self):
        f = open(self.cache_file, "wb")
        pickle.dump(self.library, f)
        f.close()


class Cache_partial:
    """ """

    def __init__(self, cls, use_cache_file=False, save_folder="."):
        self.cls = cls
        self.library = {}
        self.use_cache_file = use_cache_file
        if self.use_cache_file:
            self.cache_file = os.path.join(save_folder, cls.__name__ + "_cache.pkl")
            if os.path.isfile(self.cache_file):
                f = open(self.cache_file, "rb")
                self.library_params = pickle.load(f)
                f.close()
            else:
                self.library_params = {}
            import atexit

            atexit.register(self.save_cache)

    def __call__(self, *args):
        str_id = "".join(list(str(arg) for arg in args)) + str(env_param)
        if not str_id in self.library:
            if self.use_cache_file:
                if str_id in self.library_params:
                    params = self.library_params[str_id]
                    self.library[str_id] = self.cls(params, fast_init=True)
                else:
                    obj = self.cls(*args)
                    self.library_params[str_id] = obj.params
                    self.library[str_id] = obj
            else:
                self.library[str_id] = self.cls(*args)
        return self.library[str_id]

    def reset(self, new_save_folder=None):
        self.library = {}
        if self.use_cache_file:
            self.library_params = {}
        if new_save_folder:
            self.save_folder = new_save_folder

    def save_cache(self):
        f = open(self.cache_file, "wb")
        pickle.dump(self.library_params, f)
        f.close()



from keopscore import debug_ops_at_exec
from keopscore.binders.cpp.Cpu_link_compile import Cpu_link_compile
from keopscore.mapreduce.cpu.CpuAssignZero import CpuAssignZero
from keopscore.mapreduce.MapReduce import MapReduce
from keopscore.utils.code_gen_utils import c_include
import keopscore


class CpuReduc(MapReduce, Cpu_link_compile):
    """
    class for generating the final C++ code, Cpu version
    """

    AssignZero = CpuAssignZero

    def __init__(self, *args):
        MapReduce.__init__(self, *args)
        Cpu_link_compile.__init__(self)
        self.dimy = self.varloader.dimy

    def get_code(self):

        super().get_code()

        i = self.i
        j = self.j
        red_formula = self.red_formula
        fout = self.fout
        outi = self.outi
        acc = self.acc
        arg = self.arg
        args = self.args
        table = self.varloader.direct_table(args, i, j)
        sum_scheme = self.sum_scheme

        headers = ["cmath", "stdlib.h"]
        if keopscore.config.config.use_OpenMP:
            headers.append("omp.h")
        if debug_ops_at_exec:
            headers.append("iostream")
        self.headers += c_include(*headers)

        self.code = f"""
{self.headers}
template < typename TYPE > 
int CpuConv_{self.gencode_filename}(int nx, int ny, TYPE* out, TYPE **{arg.id}) {{
    #pragma omp parallel for
    for (int i = 0; i < nx; i++) {{
        {fout.declare()}
        {acc.declare()}
        {sum_scheme.declare_temporary_accumulator()}
        {red_formula.InitializeReduction(acc)}
        {sum_scheme.initialize_temporary_accumulator()}
        for (int j = 0; j < ny; j++) {{
            {red_formula.formula(fout,table)}
            {sum_scheme.accumulate_result(acc, fout, j)}
            {sum_scheme.periodic_accumulate_temporary(acc, j)}
        }}
        {sum_scheme.final_operation(acc)}
        {red_formula.FinalizeOutput(acc, outi, i)}
    }}
    return 0;
}}
                    """

        self.code += f"""
#include "stdarg.h"
#include <vector>

template < typename TYPE > 
int launch_keops_{self.gencode_filename}(int nx, int ny, int tagI, TYPE *out, TYPE **arg) {{
    
    if (tagI==1) {{
        int tmp = ny;
        ny = nx;
        nx = tmp;
    }}
    
    return CpuConv_{self.gencode_filename}< TYPE >(nx, ny, out, arg);

}}
template < typename TYPE >
int launch_keops_cpu_{self.gencode_filename}(int dimY,
                                             int nx,
                                             int ny,
                                             int tagI,
                                             int tagZero,
                                             int use_half,
                                             int dimred,
                                             int use_chunk_mode,
                                             std::vector< int > indsi, std::vector< int > indsj, std::vector< int > indsp,
                                             int dimout,
                                             std::vector< int > dimsx, std::vector< int > dimsy, std::vector< int > dimsp,
                                             int **ranges,
                                             std::vector< int > shapeout, TYPE *out,
                                             TYPE **arg,
                                             std::vector< std::vector< int > > argshape) {{

    
    return launch_keops_{self.gencode_filename} < TYPE >(nx, ny, tagI, out, arg);

}}
                """



from keopscore.formulas.VectorizedComplexScalarOp import VectorizedComplexScalarOp

# /////////////////////////////////////////////////////////////////////////
# ////      Conj : complex conjugate                           ////
# /////////////////////////////////////////////////////////////////////////


class Conj(VectorizedComplexScalarOp):

    string_id = "Conj"

    def ScalarOp(self, out, inF):
        string = out[0].assign(inF[0])
        string += out[1].assign(-inF[1])
        return string

    def DiffT(self, v, gradin):
        f = self.children[0]
        return f.DiffT(v, Conj(gradin))



from keopscore.utils.code_gen_utils import VectCopy
from keopscore.formulas.reductions.Min_ArgMin_Reduction_Base import (
    Min_ArgMin_Reduction_Base,
)


class Min_ArgMin_Reduction(Min_ArgMin_Reduction_Base):
    """Implements the min+argmin reduction operation : for each i or each j, find the minimal value of Fij
    and its index operation is vectorized: if Fij is vector-valued, min+argmain is computed for each dimension."""

    string_id = "Min_ArgMin_Reduction"

    def __init__(self, formula, tagIJ):
        super().__init__(formula, tagIJ)
        self.dim = 2 * formula.dim

    def FinalizeOutput(self, acc, out, i):
        return VectCopy(out, acc)



from keopscore.config.chunks import dimchunk
from keopscore.utils.code_gen_utils import GetDims, GetInds, Var_loader


class Chunk_Mode_Constants:
    def __init__(self, red_formula):

        varloader = Var_loader(red_formula)

        self.red_formula = red_formula
        self.dimred = red_formula.dimred  # dimension of reduction operation
        self.dimsp = varloader.dimsp  # dimensions of parameters variables
        self.indsp = varloader.indsp
        self.dimp = varloader.dimp
        self.dimout = (
            red_formula.dim
        )  # dimension of output variable of reduction operation
        formula = red_formula.formula
        self.dimfout = formula.dim  # dimension of output variable of inner function

        chunked_formula = formula.chunked_formulas(dimchunk)[0]
        self.dim_org = chunked_formula["dim_org"]
        self.nchunks = 1 + (self.dim_org - 1) // dimchunk
        self.dimlastchunk = self.dim_org - (self.nchunks - 1) * dimchunk
        self.nminargs = varloader.nminargs
        self.fun_chunked = chunked_formula["formula"]
        self.dimout_chunk = self.fun_chunked.dim

        self.varsi_chunked = self.fun_chunked.chunked_vars(red_formula.tagI)
        self.dimsx_chunked = GetDims(self.varsi_chunked)
        self.indsi_chunked = GetInds(self.varsi_chunked)

        self.varsj_chunked = self.fun_chunked.chunked_vars(red_formula.tagJ)
        self.dimsy_chunked = GetDims(self.varsj_chunked)
        self.indsj_chunked = GetInds(self.varsj_chunked)

        self.fun_postchunk = formula.post_chunk_formula(self.nminargs)

        self.varsi_postchunk = self.fun_postchunk.Vars(red_formula.tagI)

        self.dimsx_postchunk = GetDims(self.varsi_postchunk)
        self.indsi_postchunk = GetInds(self.varsi_postchunk)

        self.varsj_postchunk = self.fun_postchunk.Vars(red_formula.tagJ)
        self.dimsy_postchunk = GetDims(self.varsj_postchunk)
        self.indsj_postchunk = GetInds(self.varsj_postchunk)

        self.varsi_notchunked = list(
            set.union(
                set(self.varsi_postchunk),
                set(self.fun_chunked.notchunked_vars(red_formula.tagI)),
            )
        )

        # Here we detect if chunked variables are also used in the postchunk formula
        # Currently the code in GpuReduc1D_chunks.py does not handle this case, so
        # we will use the "chunk_postchunk_mix" tag defined below
        # in get_keops_dll to disable the chunked mode for the formula.
        self.chunk_postchunk_mix = (
            len(set.intersection(set(self.indsi_postchunk), set(self.indsi_chunked)))
            + len(set.intersection(set(self.indsj_postchunk), set(self.indsj_chunked)))
        ) > 0

        self.indsi_notchunked = GetInds(self.varsi_notchunked)
        self.dimsx_notchunked = GetDims(self.varsi_notchunked)
        self.dimx_notchunked = sum(self.dimsx_notchunked)

        self.varsj_notchunked = list(
            set.union(
                set(self.varsj_postchunk),
                set(self.fun_chunked.notchunked_vars(red_formula.tagJ)),
            )
        )
        self.indsj_notchunked = GetInds(self.varsj_notchunked)
        self.dimsy_notchunked = GetDims(self.varsj_notchunked)
        self.dimy_notchunked = sum(self.dimsy_notchunked)

        self.fun_lastchunked = formula.chunked_formulas(self.dimlastchunk)[0]["formula"]

        self.varsi_lastchunked = self.fun_lastchunked.chunked_vars(red_formula.tagI)
        self.indsi_lastchunked = GetInds(self.varsi_lastchunked)
        self.dimsx_lastchunked = GetDims(self.varsi_lastchunked)

        self.varsj_lastchunked = self.fun_lastchunked.chunked_vars(red_formula.tagJ)
        self.indsj_lastchunked = GetInds(self.varsj_lastchunked)
        self.dimsy_lastchunked = GetDims(self.varsj_lastchunked)

        self.varsi = [*self.varsi_notchunked, *self.varsi_chunked]
        self.dimsx = GetDims(self.varsi)
        self.indsi = GetInds(self.varsi)
        self.dimx = sum(self.dimsx)

        self.varsj = [*self.varsj_notchunked, *self.varsj_chunked]
        self.dimsy = GetDims(self.varsj)
        self.indsj = GetInds(self.varsj)
        self.dimy = sum(self.dimsy)

        self.inds = [*self.indsi, *self.indsj, *self.indsp]

        self.varsi_last = [*self.varsi_notchunked, *self.varsi_lastchunked]
        self.indsi_last = GetInds(self.varsi_last)
        self.dimsx_last = GetDims(self.varsi_last)

        self.varsj_last = [*self.varsj_notchunked, *self.varsj_lastchunked]
        self.indsj_last = GetInds(self.varsj_last)
        self.dimsy_last = GetDims(self.varsj_last)



from keopscore.formulas.Operation import Operation
from keopscore.utils.code_gen_utils import c_for_loop
from keopscore.utils.math_functions import keops_sincos
from keopscore.formulas.complex.Real2Complex import Real2Complex
from keopscore.formulas.complex.Imag2Complex import Imag2Complex
from keopscore.formulas.maths.Cos import Cos
from keopscore.formulas.maths.Sin import Sin

# /////////////////////////////////////////////////////////////////////////
# ////      ComplexExp1j                           ////
# /////////////////////////////////////////////////////////////////////////


class ComplexExp1j(Operation):

    string_id = "ComplexExp1j"

    def __init__(self, f):
        self.dim = 2 * f.dim
        super().__init__(f)

    def Op(self, out, table, inF):
        forloop, i = c_for_loop(0, self.dim, 2, pragma_unroll=True)
        body = keops_sincos(inF[i / 2], out[i] + 1, out[i])
        return forloop(body)

    def DiffT(self, v, gradin):
        # building equivalent formula for autodiff
        f = self.children[0]
        AltFormula = Real2Complex(Cos(f)) + Imag2Complex(Sin(f))
        return AltFormula.DiffT(v, gradin)



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.utils.math_functions import keops_atan2


# //////////////////////////////////////////////////////////////
# ////                 ATAN2 :  Atan2< F, G >               ////
# //////////////////////////////////////////////////////////////


class Atan2(VectorizedScalarOp):
    string_id = "Atan2"

    ScalarOpFun = keops_atan2

    @staticmethod
    def Derivative(f, g):
        r2 = f**2 + g**2
        return g / r2, -f / r2



from keopscore.utils.code_gen_utils import (
    c_zero_float,
    c_for_loop,
    c_variable,
    new_c_varname,
)
from keopscore.formulas.reductions.Reduction import Reduction


class Sum_Reduction(Reduction):
    """Sum reduction class"""

    string_id = "Sum_Reduction"

    def __init__(self, formula, tagIJ):
        super().__init__(formula, tagIJ)
        self.dim = formula.dim  # dimension of final output of reduction
        self.dimred = self.dim  # dimension of inner reduction variables
        self.dim_kahan = self.dim

    def InitializeReduction(self, tmp):
        # Returns C++ code to be used at initialization phase of the reduction.
        # Here it consists in setting the output array to zero.
        return tmp.assign(c_zero_float)

    def ReducePairScalar(self, tmp, xi):
        # Subroutine of ReducePairShort and ReducePair methods.
        # Returns C++ code that implements the "+=" accumulation operation of the sum reduction
        return tmp.add_assign(xi)

    def KahanScheme(self, acc, xi, tmp):
        loop, k = c_for_loop(0, self.dim, 1, pragma_unroll=True)
        a = c_variable(acc.dtype, new_c_varname("a"))
        b = c_variable(acc.dtype, new_c_varname("b"))
        return loop(
            a.declare_assign(xi[k] - tmp[k])
            + b.declare_assign(acc[k] + a)
            + tmp[k].assign((b - acc[k]) - a)
            + acc[k].assign(b)
        )

    def DiffT(self, v, gradin, f0=None):
        from keopscore.formulas.autodiff import Grad

        return Sum_Reduction(Grad(self.formula, v, gradin), v.cat % 2)



from keopscore.formulas.Operation import Operation
from keopscore.utils.code_gen_utils import use_pragma_unroll

####################################
######  Tensor Dot Product     #####
####################################


def prod(x):
    # product of all elements in list of integers
    res = 1
    for item in x:
        res *= item
    return res


def select(x, ind):
    # indexing of list via list of integers
    return [x[i] for i in ind]


def delete(x, ind):
    # delete items given by indices in list
    n = len(x)
    indkeep = list(set(range(n)) - set(ind))
    indkeep.sort()
    return select(x, indkeep)


def cumprod_array(x):
    # special cumulative product
    if len(x) == 0:
        return x
    else:

        def cumprod(x):
            res = x.copy()
            for i in range(1, len(x)):
                res[i] *= res[i - 1]
            return res

        return cumprod(x[1:][::-1])[::-1] + [1]


def permutation(perm, arr):
    if perm is None:
        return arr
    else:
        tmp = sorted(range(len(perm)), key=perm.__getitem__)
        return select(arr, tmp)


class TensorDot(Operation):
    string_id = "TensorDot"

    def __init__(self, fa, fb, dimsfa, dimsfb, contfa, contfb, permute=None):

        dimsfa = list(dimsfa)
        dimsfb = list(dimsfb)
        contfa = list(contfa)
        contfb = list(contfb)

        assert select(dimsfb, contfb) == select(dimsfa, contfa)

        assert fa.dim == prod(dimsfa)
        assert fb.dim == prod(dimsfb)

        super().__init__(fa, fb)

        self.dimfa = dimsfa
        self.dimfb = dimsfb
        self.contdims = select(dimsfa, contfa)

        self.indices_keepdim_a = delete(list(range(len(dimsfa))), contfa)
        self.keepdims_a = delete(dimsfa, contfa)
        self.contdims_a = select(dimsfa, contfa)
        self.list_strides_dimsfa = cumprod_array(dimsfa)

        self.indices_keepdim_b = delete(list(range(len(dimsfb))), contfb)
        self.keepdims_b = delete(dimsfb, contfb)
        self.contdims_b = select(dimsfb, contfb)
        self.list_strides_dimsfb = cumprod_array(dimsfb)

        self.keepdims = self.keepdims_a + self.keepdims_b
        self.list_strides_keepdim = cumprod_array(permutation(permute, self.keepdims))

        self.dim = fa.dim * fb.dim
        self.dim = int(self.dim / prod(self.contdims) ** 2) if len(contfa) else self.dim

        if permute is None:
            permute = list(range(len(self.keepdims)))
        else:
            assert permutation(permute, permute) == list(range(len(self.keepdims)))

        self.permute = permute

        # loop
        self.loopdim = self.keepdims + self.contdims_a
        self.dimloop = prod(self.loopdim)
        self.number_of_dimloop = len(dimsfa) + len(dimsfb) - len(contfa)

        self.ala = list(range(len(self.keepdims_a))) + list(
            range(len(self.keepdims), self.number_of_dimloop)
        )

        self.ali = self.indices_keepdim_a + contfa
        self.list_indices_a_intot = permutation(self.ali, self.ala)

        self.bla = list(range(len(self.keepdims_a), len(self.keepdims))) + list(
            range(len(self.keepdims), self.number_of_dimloop)
        )

        self.bli = self.indices_keepdim_b + contfb
        self.list_indices_b_intot = permutation(self.bli, self.bla)

        # Gradient
        self.dimfa_grad = permutation(permute, self.keepdims)

        self.list_indices_keepdim_a_inout = list(range(0, len(self.keepdims_a)))
        self.reordered_contfa = permutation(contfb, contfa)
        self.reordered_keepdim_a = permutation(
            select(permute, self.list_indices_keepdim_a_inout), self.indices_keepdim_a
        )
        self.moveaxis_a = self.reordered_keepdim_a + self.reordered_contfa

        self.list_indices_keepdim_b_inout = list(
            range(len(self.keepdims_a), len(self.keepdims))
        )
        self.reordered_contfb = permutation(contfa, contfb)
        self.reordered_keepdim_b = permutation(
            select(permute, self.list_indices_keepdim_b_inout), self.indices_keepdim_b
        )
        self.moveaxis_b = self.reordered_keepdim_b + self.reordered_contfb

        self.contfa_grad = select(permute, self.list_indices_keepdim_b_inout)
        self.contfb_grad = select(permute, self.list_indices_keepdim_a_inout)

    def Op(self, out, table, arg0, arg1):
        # returns the atomic piece of c++ code to evaluate the function on arg and return
        # the result in out

        str_code = ""

        for i in range(len(self.loopdim)):
            str_code += (
                f"for(int TD_var_{chr(70 + i)}=0; TD_var_{chr(70 + i)}<{self.loopdim[i]}; ++TD_var_{chr(70 + i)})"
                + "{\n"
                + i * "    "
            )

        list_indices_keepdim = permutation(self.permute, range(len(self.keepdims)))
        str_out_indices = ""
        for i, v in enumerate(list_indices_keepdim):
            str_out_indices += (
                f"TD_var_{chr(70 + v)} * {self.list_strides_keepdim[i]} + "
            )

        str_a_indices = ""
        for i, v in enumerate(self.list_indices_a_intot):
            str_a_indices += f"TD_var_{chr(70 + v)} * {self.list_strides_dimsfa[i]} + "

        str_b_indices = ""
        for i, v in enumerate(self.list_indices_b_intot):
            str_b_indices += f"TD_var_{chr(70 + v)} * {self.list_strides_dimsfb[i]} + "

        str_code += (
            len(self.loopdim) * "    "
            + f"{out.id}[{str_out_indices[:-2]}] += {arg0.id}[{str_a_indices[:-2]}] * {arg1.id}[{str_b_indices[:-2]}];\n"
        )

        str_code += len(self.loopdim) * "}\n"

        return f"""
                    #if C_CONTIGUOUS     // row major
                        {use_pragma_unroll()}
                        for (int i = 0; i < {out.dim}; i++)
                            {out.id}[i] = ({out.dtype})(0.0f);
                        
                        {use_pragma_unroll()}                       
                        {str_code}
                    #else               // column major
                        
                    #endif
                """

    def DiffT(self, v, gradin):
        f = self.children[0]
        g = self.children[1]
        return f.DiffT(
            v,
            TensorDot(
                gradin,
                g,
                self.dimfa_grad,
                self.dimfb,
                self.contfa_grad,
                self.indices_keepdim_b,
                self.moveaxis_a,
            ),
        ) + g.DiffT(
            v,
            TensorDot(
                gradin,
                f,
                self.dimfa_grad,
                self.dimfa,
                self.contfb_grad,
                self.indices_keepdim_a,
                self.moveaxis_b,
            ),
        )



from keopscore.binders.nvrtc.Gpu_link_compile import Gpu_link_compile
from keopscore.mapreduce.MapReduce import MapReduce
from keopscore.utils.code_gen_utils import (
    c_include,
    c_zero_float,
)


class GpuAssignZero(MapReduce, Gpu_link_compile):
    # class for generating the final C++ code, Gpu version

    def __init__(self, *args):
        MapReduce.__init__(self, *args)
        Gpu_link_compile.__init__(self)
        self.dimy = self.varloader.dimy

    def get_code(self):

        super().get_code()

        outi = self.outi
        dtype = self.dtype
        arg = self.arg
        varloader = self.varloader

        if dtype == "half2":
            self.headers += c_include("cuda_fp16.h")

        self.code = f"""
                        {self.headers}

                        extern "C" __global__ void GpuConv1DOnDevice(int nx, int ny, {dtype} *out, {dtype} **{arg.id}) {{
    
                          // get the index of the current thread
                          int i = blockIdx.x * blockDim.x + threadIdx.x;

                          if (i < nx) {{
                            {outi.assign(c_zero_float)}
                          }}

                        }}
                    """



"""
This is the main entry point for all binders. It takes as inputs :
  - map_reduce_id : string naming the type of map-reduce scheme to be used : either "CpuReduc", "GpuReduc1D_FromDevice", ...
  - red_formula_string : string expressing the formula, such as "Sum_Reduction((Exp(Minus(Sum(Square((Var(0,3,0) / Var(1,3,1)))))) * Var(2,1,1)),0)",
  - enable_chunks : -1, 0 or 1, for Gpu mode only, enable special routines for high dimensions (-1 means automatic setting)
  - enable_finalchunks : -1, 0 or 1, for Gpu mode only, enable special routines for final operation in high dimensions (-1 means automatic setting)
  - mul_var_highdim : -1, 0 or 1, for Gpu mode only, another option for special routines of final operation in high dimensions (-1 means automatic setting)
  - aliases : list of strings expressing the aliases list, which may be empty,
  - nargs : integer specifying the number of arguments for the call to the routine,
  - dtype : string specifying the float type of the arguments  "float", "double" or "half")
  - dtypeacc : string specifying the float type of the accumulator of the reduction ("float", "double" or "half")
  - sum_scheme_string : string specifying the type of accumulation for summation reductions : either "direct_sum", "block_sum" or "kahan_scheme".
  - tagHostDevice : 0 or 1, for Gpu mode only, use indicates whether data is stored on Host (0) or Gpu Device (1)
  - tagCPUGPU : 0 or 1, indicates whether we use Cpu (0) or Gpu (1) mode, i.e. reduction is performed on Cpu or Gpu
  - tag1D2D : 0 or 1, for Gpu mode only, use 1D (0) or 2D (1) computation map-reduce scheme
  - use_half : 0 or 1, for Gpu mode only, enable special routines for half-precision data type
  - device_id : integer, for Gpu mode only, id of Gpu device to build the code for

It returns
      - tag : string, hash code used as id for the input formula and parameters
      - source_file : string, either :
            - in CPU mode : name of the source file to be compiled
            - in GPU mode : name of the main dll to be called
      - low_level_code_file : string, either :
            - in CPU mode : the empty string ""
            - in GPU mode : name of the low level code (ptx) or binary file (cubin) to be passed to the main dll.
      - tagI : integer, 0 or 1, specifying if reduction must be performed over i or j indices,
      - tagZero : integer, 0 or 1, specifying if reduction just consists in filling output with zeros,
      - use_half : 0 or 1, enable special routines for half-precision data type,
      - cuda_block_size : integer, prefered block size for Gpu kernel
      - use_chunk_mode : 0, 1 or 2, if 1 or 2, enables special routines for high dimensions,
      - tag1D2D : same as input
      - dimred : integer, dimension of the inner reduction operation.
      - dim : integer, dimension of the output tensor.
      - dimy : integer, total dimension of the j indexed variables.
      - indsi : list of integers, indices of i indexed variables.
      - indsj : list of integers, indices of j indexed variables.
      - indsp : list of integers, indices of parameter variables.
      - dimsx : list of integers, dimensions of i indexed variables.
      - dimsy : list of integers, dimensions of j indexed variables.
      - indsp : list of integers, dimensions of parameter variables.

It can be used as a Python function or as a standalone Python script (in which case it prints the outputs):
  - example (as Python function) :
      get_keops_dll("CpuReduc", "Sum_Reduction((Exp(Minus(Sum(Square((Var(0,3,0) / Var(1,3,1)))))) * Var(2,1,1)),0)", 0, 0, 0, [], 3, "float", "float", "block_sum", 0, 0, 0, 0, 0)
  - example (as Python script) :
      python get_keops_dll.py CpuReduc "Sum_Reduction((Exp(Minus(Sum(Square((Var(0,3,0) / Var(1,3,1)))))) * Var(2,1,1)),0)" 0 0 0 "[]" 3 float float block_sum 0 0 0 0 0
"""
import inspect
import sys

import keopscore.config.config
from keopscore.config.config import get_build_folder
import keopscore.mapreduce
from keopscore import cuda_block_size
from keopscore.config.chunks import (
    get_enable_chunk,
    set_enable_chunk,
    dimchunk,
    set_enable_finalchunk,
    use_final_chunks,
    set_mult_var_highdim,
)
from keopscore.formulas import Zero_Reduction, Sum_Reduction
from keopscore.formulas.GetReduction import GetReduction
from keopscore.formulas.variables.Zero import Zero
from keopscore.utils.Cache import Cache
from keopscore.utils.code_gen_utils import KeOps_Error

# Get every classes in mapreduce
map_reduce = dict(inspect.getmembers(keopscore.mapreduce, inspect.isclass))


def get_keops_dll_impl(
    map_reduce_id,
    red_formula_string,
    enable_chunks,
    enable_finalchunks,
    mul_var_highdim,
    aliases,
    *args,
):
    # detecting the need for special chunked computation modes :
    use_chunk_mode = 0
    if "Gpu" in map_reduce_id:
        if not keopscore.config.config.use_cuda:
            KeOps_Error(
                "You selected a Gpu reduce scheme but KeOps is in Cpu only mode."
            )
        set_enable_chunk(enable_chunks)
        set_enable_finalchunk(enable_finalchunks)
        set_mult_var_highdim(mul_var_highdim)
        red_formula = GetReduction(red_formula_string, aliases)
        if use_final_chunks(red_formula) and map_reduce_id != "GpuReduc2D":
            use_chunk_mode = 2
            map_reduce_id += "_finalchunks"
        elif get_enable_chunk() and map_reduce_id != "GpuReduc2D":
            if len(red_formula.formula.chunked_formulas(dimchunk)) == 1:
                from keopscore.mapreduce.Chunk_Mode_Constants import (
                    Chunk_Mode_Constants,
                )

                chk = Chunk_Mode_Constants(red_formula)
                if not chk.chunk_postchunk_mix:
                    use_chunk_mode = 1
                    map_reduce_id += "_chunks"
    # Instantiation of
    map_reduce_class = map_reduce[map_reduce_id]

    map_reduce_obj = map_reduce_class(red_formula_string, aliases, *args)

    # detecting the case of formula being equal to zero, to bypass reduction.
    rf = map_reduce_obj.red_formula
    if isinstance(rf, Zero_Reduction) or (
        isinstance(rf.formula, Zero) and isinstance(rf, Sum_Reduction)
    ):
        if "Gpu" in map_reduce_id:
            map_reduce_class = map_reduce["GpuReduc1D"]
        map_reduce_obj = map_reduce_class.AssignZero(red_formula_string, aliases, *args)
        tagZero = 1
    else:
        tagZero = 0

    res = map_reduce_obj.get_dll_and_params()

    tag1D2D = 0 if tagZero == 1 else res["tag1D2D"]

    return (
        res["tag"],
        res["source_file"],
        res["low_level_code_file"],
        res["tagI"],
        tagZero,
        res["use_half"],
        cuda_block_size,
        use_chunk_mode,
        tag1D2D,
        res["dimred"],
        res["dim"],
        res["dimy"],
        res["indsi"],
        res["indsj"],
        res["indsp"],
        res["dimsx"],
        res["dimsy"],
        res["dimsp"],
    )


get_keops_dll = Cache(
    get_keops_dll_impl,
    use_cache_file=True,
    save_folder=get_build_folder(),
)


if __name__ == "__main__":
    argv = sys.argv[1:]

    argdict = {
        "map_reduce_id": str,
        "red_formula_string": str,
        "enable_chunks": int,
        "enable_finalchunks": int,
        "mul_var_highdim": int,
        "aliases": list,
        "nargs": int,
        "dtype": str,
        "dtypeacc": str,
        "sum_scheme_string": str,
        "tagHostDevice": int,
        "tagCPUGPU": int,
        "tag1D2D": int,
        "use_half": int,
        "device_id": int,
    }

    if len(argv) != len(argdict):
        KeOps_Error(
            f"Invalid call to Python script {sys.argv[0]}. There should be {len(argdict)} arguments corresponding to:\n{list(argdict.keys())}"
        )

    for k, key in enumerate(argdict):
        argtype = argdict[key]
        argval = argv[k] if argtype == str else eval(argv[k])
        if not isinstance(argval, argtype):
            KeOps_Error(
                f"Invalid call to Python script {sys.argv[0]}. Argument number {k + 1} ({key}) should be of type {argtype} but is of type {type(argval)}"
            )
        argdict[key] = argval

    res = get_keops_dll(argdict["map_reduce_id"], *list(argdict.values())[1:])
    for item in res:
        print(item)



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.formulas.variables.Zero import Zero
from keopscore.utils.math_functions import keops_floor


class Floor(VectorizedScalarOp):
    """the floor vectorized operation"""

    string_id = "Floor"

    ScalarOpFun = keops_floor

    def DiffT(self, v, gradin):
        return Zero(v.dim)
